/* bind.c */

/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */


/*****************************************************************************

  bind.c -

*****************************************************************************/

#include <stdio.h>
#include <isode/quipu/common.h>
#include <isode/quipu/bind.h>
#include <isode/quipu/dap2.h>
#include <isode/logger.h>
#include "bind.h"
#include "util.h"

extern char *dsa_address;
extern char *backup_dsa_address;
extern char *username;
extern char *password;
extern char callingDteNumber[];

extern int boundToDSA;
extern int bulkloadLogLevel;

extern LLog *bulkload_log;

/* this is not the way to do it */
extern int dsap_ad;

struct DAPconnect         dc;
struct DAPindication      di;

static int bindres;
static int triedBackup = FALSE;
static int bindStarted = FALSE;
static int boundOnce = FALSE;

int tryBackup()
{

  if (triedBackup == TRUE)
    return NOTOK;
  if (backup_dsa_address == NULLCP)
    return NOTOK;
  else
  {
    free(dsa_address);
    dsa_address = copy_string(backup_dsa_address);
    triedBackup = TRUE;
    return OK;
  }
}

/*
 * bind_to_ds - Bind to directory
 */

int init_bind_to_ds(assoc)
int * assoc;
{
struct ds_bind_arg bind_arg;
struct PSAPaddr             *addr;
void exit();

#ifdef DEBUG
fprintf( stderr, "DSA address %s \n", dsa_address );
#endif

  if (dsa_address == NULLCP)
  {
    (void) fprintf(stderr, "No dsa address has been configured in dsaptailor or detailor\n\n");
    exit(-1);
  }

  if (triedBackup == FALSE)
    if (boundOnce == FALSE)
      (void)printf("Connecting to the Directory - wait just a moment please ...\n");

  bind_arg.dba_version   = DBA_VERSION_V1988;
  bind_arg.dba_auth_type = DBA_AUTH_SIMPLE;

  bind_arg.dba_dn = str2dn(username);

  if (password == NULLCP)
  {
    bind_arg.dba_passwd[0] = '\0';
    bind_arg.dba_passwd_len = 0;
  }
  else
  {
    (void)strcpy(bind_arg.dba_passwd, password);
    bind_arg.dba_passwd_len = strlen(password);
  }

try_bind:

  if ((addr = str2paddr (dsa_address)) == NULLPA) {
    (void) fprintf(stderr, "DSA address format problem\n");
    exit(-1);
  }

  bindres = DapAsynBindRequest(addr, &bind_arg, &dc, &di, ROS_ASYNC);

  if (bindres == NOTOK)
  {
    (void) fprintf(stderr, "\n\nDirectory server temporarily unavailable\n\n");
    if (tryBackup() == OK)
    {
      (void) fprintf(stderr, "Trying another server ...\n\n");
      goto try_bind;
    }

    if( bind_arg.dba_dn != NULLDN )
      dn_free( bind_arg.dba_dn ) ;

    return NOTOK;
  }
  
  *assoc = dc.dc_sd;
  dsap_ad = dc.dc_sd;

  if( bind_arg.dba_dn != NULLDN )
    dn_free( bind_arg.dba_dn ) ;

  return OK;
}


/*
 * Attempt to bind to a referral directory..
 */

int initBindToDSonReferral( referralPSAP, assocp )
struct	PSAPaddr * referralPSAP ;
int	* assocp ;
{
struct ds_bind_arg bind_arg;
void exit();

	(void)printf("Connecting to the referral Directory - wait just a moment please ...\n");

	bind_arg.dba_version   = DBA_VERSION_V1988;
	bind_arg.dba_auth_type = DBA_AUTH_SIMPLE;

	bind_arg.dba_dn = str2dn( username );

	if( password == NULLCP )
	{
	    bind_arg.dba_passwd[0] = '\0';
	    bind_arg.dba_passwd_len = 0;
	}
	else
	{
	    (void)strcpy(bind_arg.dba_passwd, password);
	    bind_arg.dba_passwd_len = strlen(password);
	}

	bindres = DapAsynBindRequest( referralPSAP, &bind_arg, &dc, &di, ROS_ASYNC);

	if (bindres == NOTOK)
	{
	    (void) fprintf(stderr, "\n\nDirectory server temporarily unavailable\n\n");

	    return NOTOK;
	}
  
	dsap_ad = dc.dc_sd;
  	*assocp = dc.dc_sd;

	return OK;
}


int wait_bind_to_ds( assoc, wantToBlock )
int assoc;
int wantToBlock;
{
  struct PSAPindication   pi_s;
  struct PSAPindication   * pi = &(pi_s);
  int     nfds, nevents;
  fd_set  rfds;
  fd_set  wfds;

  while ((bindres == CONNECTING_1) || (bindres == CONNECTING_2)) 
  {

    FD_ZERO (&rfds);
    FD_ZERO (&wfds);

    if ((bindres == CONNECTING_2) && (wantToBlock == FALSE))
      return CONNECTING_2;

    if (bindres == CONNECTING_1) 
    {
      if (PSelectMask (assoc, &wfds, &nfds, pi) == NOTOK) 
      {
         (void) fprintf(stderr, "PSelectMask (write) failed\n");
         return NOTOK;
      }
    }

    if (bindres == CONNECTING_2) 
    {
      if (PSelectMask (assoc, &rfds, &nfds, pi) == NOTOK) 
      {
        (void) fprintf(stderr, "PSelectMask (read) failed\n");
        return NOTOK;
      }
    }

    nevents = xselect (nfds, &rfds, &wfds, NULLFD, NOTOK);
  
    if (nevents == NOTOK) 
    {
      (void) fprintf(stderr, "xselect failed\n");
      return NOTOK;
    } 
    else 
    {
      /*
       * Hack coming up: nevents is number of events, we only asked for
       * one event, so unless there are no events we go ahead without
       * checking those horrid fd things, which would be safer.
       */
       if (nevents != OK) 
         bindres = DapAsynBindRetry(assoc, 0, &dc, &di);
    }
  
    if (bindres == DONE)
    {
      if (dc.dc_result != DC_RESULT)
      {
        if (dc.dc_un.dc_bind_err.dbe_type == DBE_TYPE_SECURITY)
 	{
	  if (dc.dc_un.dc_bind_err.dbe_value == DSE_SC_INVALIDCREDENTIALS)
	  {
	    (void)fprintf(stderr, "\nBind error - check the username/password in the tailor file\n");
	  }
	  else if (dc.dc_un.dc_bind_err.dbe_value == DSE_SC_AUTHENTICATION)
            (void)fprintf(stderr, "\nBind error - alert system administrator\n");
	}
        else
	  (void)fprintf(stderr, "\nBind error - type %d, code %d\n", 
	    dc.dc_un.dc_bind_err.dbe_type, dc.dc_un.dc_bind_err.dbe_value);
        return NOTOK;
      }
      boundToDSA = TRUE;
      boundOnce  = TRUE;

      if (bulkloadLogLevel)
        (void) ll_log (bulkload_log, LLOG_NOTICE, NULLCP, "Bound: %s", callingDteNumber);

      return OK;
    }
  }

  if (tryBackup() == OK)
    if (init_bind_to_ds(&assoc) == OK)
      return OK;

  (void)fprintf(stderr, "\nCouldn't bind to the DSA - probably something wrong with the DSA address\n\n");

  return NOTOK; /* exit this way if can't talk to access point(s) */

} /* bind_to_ds */

int
rebind()
{
  if (boundToDSA == FALSE)
    return(de_bind(TRUE));
  else
    return OK;
}

int 
de_bind(wantToBlock)
int wantToBlock;
{
static int assoc;

  if (bindStarted == FALSE)
  {
    if (init_bind_to_ds(&assoc) != OK)
      return NOTOK;
    bindStarted = TRUE;
    if (wantToBlock == FALSE)
      return OK;
  }

  if (wait_bind_to_ds(assoc, wantToBlock) == NOTOK) /* don`t block */
    return NOTOK;

  return OK;
}


/*
 * Blocking referral bind..
 */ 

int 
de_referral_bind( referralPSAP )
struct	PSAPaddr * referralPSAP ;
{
static int assoc;

	if( initBindToDSonReferral( referralPSAP, &assoc ) != OK)
	    return NOTOK;
      
  	if (wait_bind_to_ds( assoc, TRUE ) == NOTOK) /* don`t block */
	    return NOTOK;
 
	return OK;
}

de_unbind()
{
  if (bulkloadLogLevel)
    (void) ll_log (bulkload_log, LLOG_NOTICE, NULLCP, "Unbind:");

  (void)ds_unbind();

  boundToDSA = FALSE;

  bindStarted = FALSE;

  return ;
}
