#include <string.h>

#include <isode/quipu/util.h>
#include <isode/quipu/dua.h>
#include <isode/quipu/ds_error.h>
#include <isode/quipu/modify.h>

#include "util.h"
#include "bulkloadglobal.h"
#include "processandupdate.h"

#define ERRMAXATTRERROR 	7
char * attributeErrorMsg[ ERRMAXATTRERROR ] =
			      { "UNKNOWN ATTRIBUTE ERROR...",
				"No such attribute...",
				"Invalid attribute syntax...",
				"Undefined attribute type...",
				"Inappropriate matching...",
				"Constraint violation...",
				"Type or value already exists..."
				};

#define ERRMSGSTORELEN	5000


char 	commentHeader[] = { "# " } ;
char 	returnTail[] = { "\n" } ;

/*
 * Returned error messages do not have comment headers on the first line
 * nor returns on the last.
 */

void
appendErrMsg( errorMsgStore, appendMsgcp )
char 	errorMsgStore[] ;
char *	appendMsgcp ;
{
int	storelen ;
char *	cp ;

	storelen = strlen( errorMsgStore ) ;

	if( ( storelen + strlen( appendMsgcp ) + 5 ) < ERRMSGSTORELEN )
	{
	    cp = errorMsgStore + storelen ;

	    if( storelen == 0 )
	    	sprintf( cp, "%s", appendMsgcp ) ;
	    else
	    	sprintf( cp, "%s%s%s", returnTail, commentHeader, appendMsgcp ) ;
	}

	return ;
}


void
errorClassTypeAndMsg( DSErrp, classp, typep, errorEntryMsgcpp, errorAttrMsgcpp, actionDefByAttrLine, EMmodSent, 
				at, av, isMasterSource, errorMatchedOKp )
struct	DSError	* DSErrp ;
int * 	classp, * typep ;
char **	errorEntryMsgcpp, ** errorAttrMsgcpp ;
int	actionDefByAttrLine, EMmodSent ;
AttributeType  at ; /* may be NULLAttrT */
AttributeValue av ; /* may be NULLAttrV */
int	isMasterSource ; /* For an attribute error case */
int *	errorMatchedOKp ;/* Assumed points to a variable set to FALSE */
{
static	char 	errorMsgStore[ ERRMSGSTORELEN ] ;
struct	DSE_at_problem * atProblemp ,
		       * best_at_problem_matchp = ((struct DSE_at_problem * ) 0) ;
#ifdef DEBUG
char * 	cp ;
#endif
        errorMsgStore[ 0 ] = '\0' ;

	*typep = *classp = 0 ;

	if( DSErrp != NULL )
	{
	    *classp = DSErrp -> dse_type ;


	    if( *classp != DSE_ATTRIBUTEERROR )
	    	*errorEntryMsgcpp = error2pstr( DSErrp ) ;

	    switch( *classp )
	    {
	    	case DSE_ATTRIBUTEERROR :

			/* Find the details for the specifed type and value .. */
			if( at != NULLAttrT )
			{
			    /* Try for a type and value match. */

			    for( atProblemp = &(DSErrp -> dse_un.dse_un_attribute.DSE_at_plist) ;
				 atProblemp != DSE_AT_NOPROBLEM && at != NULLAttrT ;
				 atProblemp = atProblemp -> dse_at_next )
			    {
			    	if( atProblemp -> DSE_at_what == DSE_AT_CONSTRAINTVIOLATION )
				{
				    /* Error can not be assigned to an individual attribute
				       as the DSE_at_type and DSE_at_value are NULL */
				    *errorEntryMsgcpp = error2pstr( DSErrp ) ;
				    break ;
				}
				else if( AttrT_cmp( atProblemp -> DSE_at_type, at ) == 0 )
				{
				    if( av != NULLAttrV )
				    {
				    	if( atProblemp -> DSE_at_value != NULLAttrV )
					{
				    	    if( AttrV_cmp( atProblemp -> DSE_at_value, av ) == 0 )
					    {
					    	best_at_problem_matchp = atProblemp ;
				    	    	break ;
					    }
					}
					else if( EMmodSent == EM_ADDATTRIBUTE &&
						 atProblemp -> DSE_at_what == DSE_AT_TYPEORVALUEEXISTS ||
						 EMmodSent == EM_ADDVALUES &&
						 atProblemp -> DSE_at_what == DSE_AT_NOSUCHATTRIBUTE ||
						 EMmodSent == EM_REMOVEVALUES &&
						 atProblemp -> DSE_at_what == DSE_AT_NOSUCHATTRIBUTE     )
					{
				    	    best_at_problem_matchp = atProblemp ;
				    	    break ;
					}
				    }
				    else
				    {
				    	best_at_problem_matchp = atProblemp ;
					break ;
				    }
				}
			    }


			    if( best_at_problem_matchp != ((struct DSE_at_problem * ) 0))
			    {
			    	if( isMasterSource == TRUE )
				    appendErrMsg( errorMsgStore, "**** Attribute error when master source ****" ) ;
				else
				    appendErrMsg( errorMsgStore, "**** Attribute error WHEN NOT THE MASTER SOURCE ****" ) ;


				if( actionDefByAttrLine == ADDATTRACTION && EMmodSent == EM_ADDVALUES )
				    appendErrMsg( errorMsgStore, "NOTE: Line designated an ADD ATTRIBUTE request but an ADD VALUE was initiated" ) ;

				else if( actionDefByAttrLine == ADDVALUEACTION && EMmodSent == EM_ADDATTRIBUTE )
				    appendErrMsg( errorMsgStore, "NOTE: Line designated an ADD VALUE request but an ADD ATTRIBUTE was initiated" ) ;
	
				*typep = best_at_problem_matchp -> DSE_at_what ;
				appendErrMsg( errorMsgStore, attributeErrorMsg[ *typep ] ) ;
			    }
			}

			if( errorMsgStore[0] != '\0' ) {
			    *errorAttrMsgcpp = strdup( errorMsgStore ) ;
			    *errorMatchedOKp = TRUE ;
			}

			break ;

		case DSE_NAMEERROR :
			*typep = DSErrp -> dse_un.dse_un_name.DSE_na_problem ;
			*errorMatchedOKp = TRUE ;
			break ;

		case DSE_SERVICEERROR :
			*typep = DSErrp -> dse_un.dse_un_service.DSE_sv_problem ;
			*errorMatchedOKp = TRUE ;
			break ;

		case DSE_SECURITYERROR :
			*typep = DSErrp -> dse_un.dse_un_security.DSE_sc_problem ;
			*errorMatchedOKp = TRUE ;
			break ;

		case DSE_ABANDON_FAILED :
			*typep = DSErrp -> dse_un.dse_un_abandon_fail.DSE_ab_problem ;
			*errorMatchedOKp = TRUE ;
			break ;

		case DSE_UPDATEERROR :
			*typep = DSErrp -> dse_un.dse_un_update.DSE_up_problem ;
			*errorMatchedOKp = TRUE ;
			break ;

		case DSE_REFERRAL :

			break ;

	    	default :
#ifdef DEBUG
fprintf( stderr, "Unknown or unsupported DSError dse_type %d\n", *classp ) ;
#endif
			break ;
	    }
	}

	return ;
}
void
errorClassAndType( DSErrp, em_type, classp, typep, at, av )
struct	DSError	* DSErrp ;
int	em_type ;
int * 	classp, * typep ;
AttributeType  at ;
AttributeValue av ;
{
struct	DSE_at_problem * atProblemp ;
#ifdef DEBUG
char * 	cp ;
#endif

	*typep = *classp = 0 ;
	
	
	if( DSErrp != NULL )
	{
	    *classp = DSErrp -> dse_type ;

	    switch( *classp )
	    {
	    	case DSE_ATTRIBUTEERROR :

			/* Find the details for the specifed type and value .. */
			if( at != NULLAttrT && ( em_type == EM_REMOVEATTRIBUTE || av != NULLAttrV ))
			{
			    for( atProblemp = &(DSErrp -> dse_un.dse_un_attribute.DSE_at_plist) ;
				 atProblemp != DSE_AT_NOPROBLEM ;
				 atProblemp = atProblemp -> dse_at_next )
			    {
				if( AttrT_cmp( atProblemp -> DSE_at_type, at ) == 0 )
				{ 
				    if( atProblemp -> DSE_at_value != NULLAttrV &&
				    	av != NULLAttrV )
				    {
				    	if( AttrV_cmp( atProblemp -> DSE_at_value, av ) == 0 )
					{
			    	    	    *typep = atProblemp -> DSE_at_what ;
				    	    break ;
					}
				    }
				    else
				    {
					*typep = atProblemp -> DSE_at_what ;
				    	break ;
				    }
				}
			    }
			}
			break ;
		case DSE_NAMEERROR :
			*typep = DSErrp -> dse_un.dse_un_name.DSE_na_problem ;
			break ;
		case DSE_SERVICEERROR :
			*typep = DSErrp -> dse_un.dse_un_service.DSE_sv_problem ;
			break ;
		case DSE_SECURITYERROR :
			*typep = DSErrp -> dse_un.dse_un_security.DSE_sc_problem ;
			break ;
		case DSE_ABANDON_FAILED :
			*typep = DSErrp -> dse_un.dse_un_abandon_fail.DSE_ab_problem ;
			break ;
		case DSE_UPDATEERROR :
			*typep = DSErrp -> dse_un.dse_un_update.DSE_up_problem ;
			break ;

		case DSE_REFERRAL :

	    	default :
#ifdef DEBUG
fprintf( stderr, "Unknown or unsupported DSError dse_type %d\n", *classp ) ;
#endif
			break ;
	    }
	}

	return ;
}
