/* Updates
 *
 *	30-Nov-92	bufferAttrs - logic changed to prevent NOTOKALLREAD when
 *				      atleast one attribute has already been found.
 *
 *	31-Mar-93	differAttributesFromRAAnode - prevent a delete attribute
 *				      from replacing delete value requests.
 *
 *	21-Apr-93	-s and -so options for differing secondary data sources.
 *			A check for missing initial rootedAt statements within
 *			differEntries was added.
 */

/*
 *Note 1: Two line buffers are used to facilitate the return of lines,
 * 	  that are read but not processed, to the next processing stage.
 */

/*
 * To be resolved..
 *
 * 1) differEntries why bufferRootedAt since should have exhausted both files by this point ???
 * 2) FILE attributes are not handled,
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>


#include <isode/quipu/name.h>

#define DELENTRIESNOTATTRS	1
#define	DELATTRSNOTENTRIES	2
#define DELATTRSNOTOBJCLSORENTS	3

#define MAXLINELEN		2048
#define INITLINEBUFLEN		20000	/* Default attribute line buffer size */
#define LINEBUFSIZEINC		10000   /* By how many octets the line buffer shall
					   be increased each time the buffer fills  
					   NOTE: MUST BE GREATER THAN "MAXLINELEN" 
					   since the realloc will then read in
					   the next line */

#define	SMALLBUFLEN		256

#define UNDEFLINETYPE		0

#define BLANKLINETYPE		1
#define COMMENTLINETYPE		2
#define SOURCELINETYPE		3
#define ROOTEDATLINETYPE	4
#define ATTSLINETYPE		5
#define OTHERLINETYPE		6  /* E.g. empty string "" */

#define ROOTEDATATTSORBLANKLINETYPE 7
#define ROOTEDATORATTSLINETYPE 	8


#define ADDRDNCHAR		'$'
#define DELETERDNCHAR		'!'
#define ADDATTRCHAR		'+'
#define DELETEATTRCHAR		'-'
#define ADDVALUECHAR		'>'
#define DELETEVALUECHAR		'<'


#define LINECONTCHAR		'\\'
#define MULTIATTRVALSCHAR	'&'
#define MULTIATTRPARTSCHAR	'#'
#define	ATTRTYPEDELIM		'='
#define COMMENTCHAR		'#'

#define TRUE			1
#define FALSE			0

#define NOTOKALLDATAREAD	2 /* Must not clash with OK or NOTOK from /quipu/attr.h */


#define OLDFILE			1
#define NEWFILE			2
#define BOTHFILES		3

#define	ALLNODES		-1

#define BOUNDEDBYEOF		-1


struct lineBufferStruct {

	char *  lineBufferp ;

	int 	bufferSize  ; /* In characters */
} ;

struct rootedAtDetails {

	char  * rootedAtLine ;
	long	fileOffset ;

	struct 	rootedAtDetails * nextRootedAtDetailsp ;

} ;

#define NULLRAD ((struct rootedAtDetails *) 0 )


struct attributeLine {

	char  *	attrLine ;	/* "attribute type = attribute value" */
	int	attrUsed ;
	int	nthAttrOfThisType ; /* Keeps track of the number of instances
				       of this attributes type held within this
				       attributeLine linked list upto and including 
				       the current node                             */
	AttributeValue	av ;
	AttributeType	at ;

	/* NOTE: av, at and nthAttrOfThisTyps exist solely within
	   the scope of `differAttributesFromRAAnode()'		  */
				       
	struct 	attributeLine * nextAttrLinep ;
} ;

#define NULLALP ((struct attributeLine *) 0 )

struct rootedAtAttributes {

	RDN	rdn ;
	char  * rdnLine ;

	int	attrsUsed ;

	struct 	attributeLine * attrLinep ;

	struct  rootedAtAttributes * nextRootedAtAttrsp ;
} ;

#define NULLRAA	((struct rootedAtAttributes *) 0 )


/* 
 * GLOBAL VARIABLES
 */

char	Version[] = "1.1" ;

int	totalRDNs = 0; /* for debug purposes only */




int
determineLineType( lineBufferp )
char * lineBufferp ;
{
int lineType = UNDEFLINETYPE ;

	if( isspace( *lineBufferp ) )
	    lineType = BLANKLINETYPE ;

	else if ( *lineBufferp == COMMENTCHAR )
	    lineType = COMMENTLINETYPE ;

	else if ( !strncmp( lineBufferp, "source", 6 ) )
	    lineType = SOURCELINETYPE ;

	else if ( !strncmp( lineBufferp, "rootedAt", 8 ) )
	    lineType = ROOTEDATLINETYPE ;

	else if ( strchr( lineBufferp, ATTRTYPEDELIM ) != NULLCP)
	    lineType = ATTSLINETYPE ;

	else lineType = OTHERLINETYPE ; /* E.g. empty string "" */


	return( lineType ) ; /* XXXLINETYPE */
}



int 
matchLineType( lineType, ReqLineType )
int lineType, ReqLineType ;
{
int	ret_val = FALSE ;

	
	if( ReqLineType == ROOTEDATATTSORBLANKLINETYPE )
	{
	    if( lineType == ROOTEDATLINETYPE ||
  		lineType == ATTSLINETYPE     ||
		lineType == BLANKLINETYPE      )
		    ret_val = TRUE ;
	}
	else if( ReqLineType == ROOTEDATORATTSLINETYPE )
	{
	    if( lineType == ROOTEDATLINETYPE ||
  		lineType == ATTSLINETYPE     )
		    ret_val = TRUE ;
	}
	else
	{
	    if( lineType == ReqLineType )
	    	    ret_val = TRUE ;
	}

	return( ret_val ); /* TRUE or FALSE */
}


/*
 * Load the next line that is not a comment or
 * a blank line into the buffer.
 */ 

int
skipCommentsAndBlanks( DMfp, lineBufferp )
FILE * DMfp ;
char * lineBufferp ;
{
int ret_val  = OK ;
int lineType = UNDEFLINETYPE ;

	while( !feof( DMfp ) &&
	       (lineType == UNDEFLINETYPE   || 
	        lineType == BLANKLINETYPE   ||
	        lineType == COMMENTLINETYPE   ) &&
 		ret_val  == OK )
	{
	    if( fgets( lineBufferp, MAXLINELEN, DMfp ) == NULLCP )
	    	ret_val = NOTOK ;
	    else
	    	lineType = determineLineType( lineBufferp ) ;
	}

	return( ret_val ); /* OK or NOTOK */
}


/*
 * Obtain next line that is not a comment or blank
 * and check that it is of the required type.
 */
 
int
getNextLine( DMfp, Bufferp, ReqLineType )
FILE * DMfp ;
char * Bufferp ;
int    ReqLineType ;
{
int ret_val = OK ;

	if( skipCommentsAndBlanks( DMfp, Bufferp ) == NOTOK )
	    ret_val = NOTOK ;
	else {
	    if( determineLineType( Bufferp ) != ReqLineType )
	    	ret_val = NOTOK ;
	}

	return( ret_val ); /* OK or NOTOK */
}

/*
 * strip off the attribute type (without the equals sign !)
 */

void
extractAttrType( returnAttrTypep, attrLineBuffer, returnAttrTypeLenp )
char *	returnAttrTypep, attrLineBuffer[] ;
int  *	returnAttrTypeLenp ;
{
int	i=0 ;

	while( (i < strlen( attrLineBuffer )) &&
	       (attrLineBuffer[i] != ATTRTYPEDELIM )   )
	    i++ ;

	if( attrLineBuffer[i] == ATTRTYPEDELIM )
	{
	    *returnAttrTypeLenp = i ;
	    strncpy( returnAttrTypep, attrLineBuffer, i ) ;

	    returnAttrTypep[ i ] = '\0' ;
	}
	else
	{
	    *returnAttrTypep = '\0' ;
	    *returnAttrTypeLenp = 0 ;
	}
	return ;
}

/*
 * Line continutation requires an end of line "\".."
 */
 
int
lineToBeContinued( lineBufferp )
char *	lineBufferp ;
{
int	lineContReq ;
char *  lineContcp ;

	if( (lineContcp = strrchr( lineBufferp, LINECONTCHAR )) != NULLCP &&
	    (++lineContcp) && (*lineContcp == '\n') )
	    lineContReq = TRUE ;
	else
	    lineContReq = FALSE ;

	return( lineContReq ) ;
}

/*
 * Inherited attributes are treated as a special case
 */

int
iattrLineToBeContinued( lineBuffercp, firstLine )
char *	lineBuffercp ;
int	firstLine ;
{
int	lineContReq = FALSE, attrTypeLen ;
char    attrLineType[SMALLBUFLEN] ;
AttributeType	at ;
static	AttributeType iattrat = NULLAttrT ;

	if( iattrat == NULLAttrT )
	    iattrat =  str2AttrT( "inheritedAttribute" ) ;

	if( iattrat != NULLAttrT )
	{
	    if( firstLine == TRUE )
	    {
		extractAttrType( attrLineType, lineBuffercp, &attrTypeLen ) ;
	    
		if( (at = str2AttrT( attrLineType )) != NULLAttrT )
		{
		    if( AttrT_cmp( at, iattrat ) == 0 )
			    lineContReq = TRUE ;
		    
		    AttrT_free( at ) ;
		}
		
	    } else {

		/* Subsequent line so check for solitary closing brace.. */
		if( *lineBuffercp != ')' || *(lineBuffercp+1) != '\n' )
		    lineContReq = TRUE ;
	    }
	}

	return( lineContReq ) ;
}


/*
 * Attempt to find the next line of the specified type.
 *
 * Note: The line buffer may already contain an unprocessed line.
 *
 * Multiple line shall be buffered if line continuation characters
 * are found or if an inheritedAttribute is being processed.
 */

int
getLine( DMfp, lineBufferStructp, ReqLineType )
FILE *	DMfp ;
struct	lineBufferStruct * lineBufferStructp ;
int	ReqLineType ;
{
int	ret_val  = OK ;
int	lineType = UNDEFLINETYPE ;
int	matchOK  = FALSE ;
int 	LineContinued = FALSE, IattrLineContinued = FALSE ;
char *	baseBuffercp = lineBufferStructp -> lineBufferp, 
     *	currBuffercp = lineBufferStructp -> lineBufferp ;
int	fileEOF, bufferSizeUsed = 0 ;

	/* The line buffer may already contain a line if the correct type.. */
	lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;
	
	while( !(fileEOF = feof( DMfp )) &&
 	       ret_val == OK &&
	       (  (matchOK = matchLineType( lineType, ReqLineType)) == FALSE
		  ||
	          (matchOK == TRUE && (LineContinued == TRUE || IattrLineContinued == TRUE) ))
	     )
	{
	    /* Check that the line buffer is large enough.. */

	    if( bufferSizeUsed >= (lineBufferStructp->bufferSize - MAXLINELEN))
	    {
	    	/* Request a larger buffer.. */
		if( (lineBufferStructp -> lineBufferp = 
			realloc( (char *)lineBufferStructp->lineBufferp, 
				 (unsigned)(lineBufferStructp->bufferSize + LINEBUFSIZEINC))) == NULLCP )
		{
		    fprintf( stderr, "Memory REALLOCATION failure when increaing buffer sizeto %d\n", 
		    		lineBufferStructp->bufferSize + LINEBUFSIZEINC );		    

		    ret_val = NOTOK ;
		}
		else 
		{
		    fprintf( stderr, "Realloc'ed to %d\n", lineBufferStructp->bufferSize + LINEBUFSIZEINC );

		    lineBufferStructp->bufferSize += LINEBUFSIZEINC ;

		    /* Adjust buffer pointer to point to the same
		       position in the realloc'ed buffer ..       */

		    currBuffercp = lineBufferStructp -> lineBufferp +
		    				(currBuffercp - baseBuffercp) ;

		    baseBuffercp = lineBufferStructp -> lineBufferp ;
		}
	    }


	    if( ret_val == OK )
	    {
		if( matchOK == FALSE )
		{
		    currBuffercp = lineBufferStructp -> lineBufferp ; /* Reset the line buffer ptr */

		    bufferSizeUsed = 0 ;
		}
	
		/* Read in a fresh line.. */
	    	if( fgets( currBuffercp, MAXLINELEN, DMfp ) == NULLCP )
{
#ifdef DEBUG
fprintf( stderr," FILE END WAS REACHED\n" );
#endif
	    	    ret_val = NOTOKALLDATAREAD ;
}
	    	else
		{
#ifdef DEBUG4
fprintf( stderr, "Read %s\n", currBuffercp ) ;
#endif
		    if( LineContinued == FALSE && IattrLineContinued == FALSE )
		    {
			/* ..no line continuations so far */
		    
	    	    	lineType = determineLineType( baseBuffercp ) ;

		     	/* Check if this line should be continued.. */
		     	if( lineToBeContinued( currBuffercp ) == TRUE )
		     	{
			    LineContinued  =  TRUE ;
			    bufferSizeUsed += strlen( currBuffercp ) ;
			    currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			}

			/* Check if this line contains an iattr attribute.. */
			if( lineType == ATTSLINETYPE && 
			    (IattrLineContinued = iattrLineToBeContinued( currBuffercp, TRUE )) == TRUE )
			{
			    bufferSizeUsed += strlen( currBuffercp ) ;
			    currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			}
		    }
		    else
		    {
		    	/* handle line continuation .. */
			
			if( LineContinued == TRUE )
			{
		     	    if( lineToBeContinued( currBuffercp ) == FALSE )
			    	LineContinued = FALSE ; /* Continuation finished */
			    else
			    {
			    	/* Prepare for appending the next line */
			    	bufferSizeUsed += strlen( currBuffercp ) ;
			    	currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			    }
			}
			else if( IattrLineContinued == TRUE )
			{
		     	    if( iattrLineToBeContinued( currBuffercp, FALSE ) == FALSE )
			    	IattrLineContinued = FALSE ; /* Continuation finished */
			    else
			    {
			    	/* Prepare for appending the next line */
			    	bufferSizeUsed += strlen( currBuffercp ) ;
			    	currBuffercp   =  baseBuffercp + bufferSizeUsed ;
			    }
			}
		    }
		}
	    }
	    
	} /* While loop */

	if( fileEOF || ( matchOK == FALSE && ret_val != NOTOK ) )
	    ret_val =  NOTOKALLDATAREAD ;

	return( ret_val ); /* OK, NOTOKALLDATAREAD */
}

void
removeAllWhiteSpaces( tokencp )
char *	tokencp ;
{
char * 	fromcp = tokencp, * tocp = tokencp ;

	while( fromcp && *fromcp )
	{
	    if( !isspace( *fromcp ) )
	    	*tocp++ = *fromcp ;

	    ++fromcp ;
	}

	if( *fromcp == '\0' )
	    *tocp = '\0' ;

	return ;
}

/*
 * Remove white space characters (i.e. returns, new line, form feed or vertical tab).
 * Replace multiple spaces and tabs with a single one.
 * Remove line continuations "\\n".
 *
 * Return pointer is to malloc'ed storage.
 */


char *
strDupTidyStringRemLineConts( linecp )
char *	linecp ;
{
char *  malloccp, * tocp, * fromcp ;
char	suspectedLineContChar = '\0', lastCharSpace = '\0' ;

	while( isspace( *linecp ) )
	    linecp++ ; /* Skip leading whitespaces */

	fromcp = linecp ;

	tocp = malloccp = strdup( linecp ) ;

	if( tocp != NULL )
	{
	    while( *fromcp )
	    {
	    	if( isspace( *fromcp ) )
		{
		    if( *fromcp == ' ' || *fromcp == '\t' )
		    {
		    	lastCharSpace = *fromcp ;
		    }
		    else lastCharSpace = '\0' ;
			
		    suspectedLineContChar = '\0' ;
		}
		else if( *fromcp != LINECONTCHAR )
		{
 		    /* Note that in the case of the following string
 		     * 'A T \26 T', that you can have both lastCharSpace
 		     * and suspectedLineContChar at the same time
 		     * PJSG: 4/12/93
 		     */
 		    if( lastCharSpace )
 			*tocp++ = lastCharSpace ;
 
 		    lastCharSpace = '\0' ;
 		    
		    if( suspectedLineContChar )
		    {
		    	*tocp++ = suspectedLineContChar ; /* LINECONTCHAR */
			suspectedLineContChar = '\0' ;
		    }
		    
		    *tocp++  = *fromcp ;
		}
		else 
		{
		    suspectedLineContChar = LINECONTCHAR ;
		}
		
		fromcp++ ;
	    }

	    *tocp = '\0' ; /* Null terminate the string before leaving */
	}
	else fprintf( stderr, "Memory allocation failure in `strDupTidyStringRemLineConts()`\n" ) ;

	return( malloccp ) ;
}

/*
 * Check that this line contains a value after the equals
 * as test data EDB had "acl=".
 */
 
int
value_present( lineBufferp )
char *	lineBufferp ;
{
int	val_present = FALSE ;
char *	cp ;

	if( (cp = strchr( lineBufferp, '=' ) ) != NULLCP )
	{
	    cp++ ; /* skip the '=' */
	
	    while( isspace( *cp ) )
	    	cp++ ;

	    if( *cp != '\0' && strncmp( cp, "{FILE}", 6 ) ) /* File lines have no values */
	    	val_present = TRUE ;
	}

	return( val_present ) ;
}

/***************************************************************************************/

void
free_attrLine_list( baseAttrLinep )
struct attributeLine * baseAttrLinep ;
{
struct attributeLine * currAttrLinep = baseAttrLinep,
		     * buffAttrLinep = NULLALP ;

	while( currAttrLinep != NULLALP )
	{
	    /* Free the attribute character line storage.. */
	    (void)free( (char*)(currAttrLinep -> attrLine) ) ;

	    buffAttrLinep = currAttrLinep ;
	    currAttrLinep = currAttrLinep -> nextAttrLinep ;

	    /* Free the node storage.. */
	    (void)free( (char*)( buffAttrLinep )) ;
	}
}




/*
 * Free the entire "rootedAtAttributes" linked list..
 */
 
void
freeRootAtAttrs_list( baseRAApp )
struct	rootedAtAttributes ** baseRAApp ;
{
struct 	rootedAtAttributes * currRAAp = *baseRAApp,
		           * buffRAAp = NULLRAA ;

	while( currRAAp != NULLRAA )
	{
	    if( currRAAp -> rdn != NULLRDN )
	    	rdn_free( currRAAp -> rdn ) ;
	
	    /* Free the RDN character line storage.. */
	    (void)free((char*)( currRAAp -> rdnLine) ) ;

	    /* Free the attributeLine storage.. */
	    free_attrLine_list( currRAAp -> attrLinep ) ;

	    buffRAAp = currRAAp ;
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;

	    /* Free the node storage.. */
	    (void)free((char*)( buffRAAp )) ;
	}

	/* All nodes returned.. */
	*baseRAApp = NULLRAA ;

	return ;
}



/*
 * Unlink and return the storage for the specified node..
 */

struct	rootedAtAttributes **
freeRootAtAttrs_node( baseRAApp, nodeNumToBeUnlinked )
struct	rootedAtAttributes ** baseRAApp ;
{
struct 	rootedAtAttributes * currRAAp = * baseRAApp,
		           * buffRAAp = NULLRAA,
			   * prevRAAp = NULLRAA,
			   * lastRAAp = NULLRAA ;
int	currNodeNum = 1 ;


	/* Note: Nodes number 1,2,.. */

	
	/* Find the node that is to be unlinked.. */
	while( currRAAp != NULLRAA && currNodeNum < nodeNumToBeUnlinked )
	{
	    currNodeNum++ ;

	    prevRAAp = currRAAp ;
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;
	}


	if( nodeNumToBeUnlinked == currNodeNum )
	{
	    if( currRAAp -> rdn != NULLRDN )
	    	rdn_free( currRAAp -> rdn ) ;
		
    	    /* Free the RDN character line storage.. */
	    (void)free((char*)( currRAAp -> rdnLine) ) ;

	    /* Free the attributeLine storage.. */
	    free_attrLine_list( currRAAp -> attrLinep ) ;

	    buffRAAp = currRAAp ;
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;

	    /* Free the node storage.. */
	    (void)free((char*)( buffRAAp )) ;


	    /* Link around the freed node */
	    if( prevRAAp )
	    	prevRAAp -> nextRootedAtAttrsp = currRAAp ;

	
	    if( currNodeNum == 1 )
	    {
	    	/* .. have freed the base node */
	    
  	    	*baseRAApp = currRAAp ;
	    }

	    /* Determine the last node.. */
	    if( currRAAp == NULLRAA )
	    	lastRAAp =  prevRAAp ; /* NULLRAA if sole node deleted */
	    else
	    {
		/* Find the last node */
		while( currRAAp != NULLRAA )
		{
	    	    buffRAAp = currRAAp ;
	    	    currRAAp = currRAAp -> nextRootedAtAttrsp ;
		}

		lastRAAp = buffRAAp ;
	    }
	}
	else
	{
	    /* Error condition */
	    freeRootAtAttrs_list( baseRAApp ) ;
	}


	/* Return a pointer to the last node in the list.. */
	if( lastRAAp )
	    return( &(lastRAAp -> nextRootedAtAttrsp) );
	else
	    return( baseRAApp ) ;
}



struct attributeLine *
create_attributeLine_node_from_type_value( attrTypecp, attrValuecp )
char *	attrTypecp, * attrValuecp ;
{
struct 	attributeLine * newAttributeLinep = NULLALP ;
char *  attrLinecp = NULLCP ;

	if( (attrTypecp) && (attrValuecp) &&
	    ( newAttributeLinep = (struct attributeLine *)malloc( 
					(unsigned)sizeof( struct attributeLine )) ) != NULLALP )
	{
	    /* ..single attribute value */

	    if( (attrLinecp = malloc( (unsigned)(strlen( attrTypecp ) + strlen( attrValuecp ) + 5))) != NULLCP )
	    {
		newAttributeLinep -> attrLine = attrLinecp ;

		/* Load up the attribute type attribute value pair.. */
		if( *( attrValuecp + (strlen(attrValuecp) - 1)) != '\n' )
		    sprintf( newAttributeLinep -> attrLine, "%s%c%s%c",
				 attrTypecp, ATTRTYPEDELIM, attrValuecp, '\n' ) ;

		else sprintf( newAttributeLinep -> attrLine, "%s%c%s",
				 attrTypecp, ATTRTYPEDELIM, attrValuecp ) ;
#ifdef DEBUG
fprintf( stderr, "Adding attr %s", newAttributeLinep -> attrLine ) ;
#endif
	    }
	    else newAttributeLinep -> attrLine = NULLCP ;

	    newAttributeLinep -> attrUsed = FALSE ;

	    newAttributeLinep -> nthAttrOfThisType = 0 ;

	    newAttributeLinep -> av = NULLAttrV ;

	    newAttributeLinep -> at = NULLAttrT ;
	    
	    newAttributeLinep -> nextAttrLinep = NULLALP ;
	}

	if( newAttributeLinep == NULLALP )
	    fprintf( stderr, "Memory allocation failure when creating a attributeLine node\n" );

	return( newAttributeLinep ) ;
}

void
setUpAttrTypeValue( baseAttrLinep )
struct	attributeLine * baseAttrLinep ;
{
struct	attributeLine * currALp = baseAttrLinep ;
char	attrLineType[SMALLBUFLEN] ;
int	attrLineLen ;
char *	valuecp ;
AttributeType  at ;
static	AttributeType iattrat = NULLAttrT ;

	if( iattrat == NULLAttrT )
	    iattrat =  str2AttrT( "inheritedAttribute" ) ;

	while( currALp )
	{
	    extractAttrType( attrLineType, currALp -> attrLine, &attrLineLen ) ;

	    at = str2AttrT( attrLineType ) ;

	    if( at != NULLAttrT )
	    {
		currALp -> at = at ;
	    	
		if( at ->oa_syntax != 0 &&
	    	    (valuecp = strchr( currALp -> attrLine, '=' ) ) != NULLCP )
		{
		    if( iattrat != NULLAttrT && AttrT_cmp( iattrat, at ) == 0 )
		    {
			while( isspace( *++valuecp ) )
			    /* Skip leading whitespace for multi-line inherited attributes */ ;

			valuecp = strdup( valuecp ) ;
		    }
		    else
		    {
			valuecp = strDupTidyStringRemLineConts( ++valuecp ) ;
		    }

		    if( valuecp != NULLCP ) 
		    {
#ifdef DEBUG4
fprintf( stderr, "Post Tidy %s\n", valuecp ) ;
#endif
		    	/* ...skipping '=' */
	    	    	currALp -> av = str2AttrV( valuecp, currALp -> at -> oa_syntax) ;

			if( currALp -> av == NULLAttrV )
			{
	    		    fprintf( stderr, "FATAL ERROR: invalid attribute value in %s", currALp -> attrLine ) ;
			    fprintf( stderr, "Will carry on looking for more problems but DISCARD output file\n" ) ;
			}
			
			free( valuecp ) ;
		    }
		}
	    }
	    else
	    {
	    	fprintf( stderr, "FATAL ERROR: invalid attribute type in %s", currALp -> attrLine ) ;
		fprintf( stderr, "Will carry on looking for more problems but DISCARD output file\n" ) ;
	    }
	
	    currALp = currALp -> nextAttrLinep ;
	}
}



void
clearDownAttrTypeValue( baseAttrLinep )
struct	attributeLine * baseAttrLinep ;
{
struct	attributeLine * currALp = baseAttrLinep ;

	while( currALp )
	{
	    AttrT_free( currALp -> at ) ;

	    AttrV_free( currALp -> av ) ;
	    
	
	    currALp = currALp -> nextAttrLinep ;
	}
}


/*
 * Skip a tail line continuation
 */

void
skipTailLineCont( attrValuecpp )
char **	attrValuecpp ;
{
char *  attrValuecp = *attrValuecpp ;

	/* Check to see if there is broken line character
	   before the next non whitespace character */

	while( attrValuecp &&
		(isspace( *attrValuecp ) ) )
	{
	    attrValuecp++ ;
	}

	if( *attrValuecp == LINECONTCHAR &&
	    (attrValuecp+1) && (*(attrValuecp+1) == '\n') )
	{
	    /* skip the line continuation and new line characters */
	    attrValuecp += 2 ;

	    *attrValuecpp = attrValuecp ;
	}

	return ;
}



/*
 * Create an "attributeLine" node to store the attribute type value pair.
 * The procedure must handle multi value attributes e.g. "telephoneNumber = X2345 & 3233"
 * aswell as attribute values that are broken across lines by the "\" operator.
 *
 * The second parameter will be loaded with the address of "nextAttrLinep"
 * in the final created node.
 */

struct attributeLine *
create_attributeLine_node( attrLinep, finalNextAttrLineppp )
char  * attrLinep ;
struct attributeLine *** finalNextAttrLineppp ;
{
struct 	attributeLine * newAttributeLinep = NULLALP, * currAttributeLinep = NULLALP ;
char  *	attrValcp  = NULLCP, * attrTypecp = NULLCP ;
char  * nextAttrValuecp = attrLinep ;
char	bufferedAttrType[SMALLBUFLEN] ;
int	bufferedAttrTypeLen  ;
int	AllAttrValuesProcessed = FALSE, multiAttrValues = FALSE ;


	do {
	    /* Check to see if the line contains multiple attribute values */

	    if( (attrValcp  = strchr( attrLinep, MULTIATTRVALSCHAR)) != NULLCP &&
		(attrTypecp = strchr( attrLinep, ATTRTYPEDELIM    )) != NULLCP   )
	    {
	    	extractAttrType( bufferedAttrType, attrLinep, &bufferedAttrTypeLen ) ;

		nextAttrValuecp = attrValcp + 1 ; /* skip the MULTIATTRVALUECHAR */
		
		skipTailLineCont( &nextAttrValuecp ) ;

		*attrValcp = '\0' ;	/* Null terminate the "attrType = attrValue" */

		attrValcp  = attrTypecp + 1 ;	  /* Skip the "=" */

		multiAttrValues   = TRUE ;

	    	currAttributeLinep = newAttributeLinep = 
				create_attributeLine_node_from_type_value(
							bufferedAttrType, attrValcp ) ;
	    }
	    else if( multiAttrValues == FALSE )
	    {
	    	if( ( currAttributeLinep = newAttributeLinep = 
			(struct attributeLine *)malloc( 
					(unsigned)sizeof( struct attributeLine )) )!= NULLALP )
	    	{
		    /* ..single attribute value */

		    /* Remove additional leading white spaces.. */
	    	    while( *attrLinep && isspace( *(attrLinep + 1) ) )
		    {	    	
			attrLinep++ ;
		    }

		    if( isspace( *attrLinep ) )
			*attrLinep = ' ' ; /* Ensure the leading white space is a ` ` */

		    newAttributeLinep -> attrLine = strdup( attrLinep ) ;
#ifdef DEBUG
fprintf( stderr, "Adding attr %s", attrLinep ) ;
#endif
		    newAttributeLinep -> attrUsed = FALSE ;

		    newAttributeLinep -> nthAttrOfThisType = 0 ;
		    
		    newAttributeLinep -> av = NULLAttrV ;

		    newAttributeLinep -> at = NULLAttrT ;

		    newAttributeLinep -> nextAttrLinep = NULLALP ;
		}

		AllAttrValuesProcessed = TRUE ;
	    }
	    else /* multiAttrValues == TRUE */
	    {
	    	if( (attrValcp  = strchr( nextAttrValuecp, MULTIATTRVALSCHAR)) != NULLCP )
		{
	    	    /* Add the next attribute value to the linked list.. */

		    *attrValcp = '\0' ; /* Null terminate the attribute value */

		    if( (currAttributeLinep -> nextAttrLinep =
		    	create_attributeLine_node_from_type_value(
					bufferedAttrType, nextAttrValuecp ) ) == NULLALP )
		    	AllAttrValuesProcessed = TRUE ;
		    else
		    {
		    	nextAttrValuecp = attrValcp + 1 ; /*Skip the position the held
		    					    the "&" before being replaced
							    by the null terminator */

			skipTailLineCont( &nextAttrValuecp ) ;

			currAttributeLinep = currAttributeLinep -> nextAttrLinep ;		    		
		    }
		}
		else 
		{
	    	    /* Add the final attribute value to the linked list.. */
#ifdef DEBUG4
fprintf( stderr, "Final attribute value %s\n", nextAttrValuecp ) ;
#endif
		    currAttributeLinep -> nextAttrLinep =
			create_attributeLine_node_from_type_value(
						bufferedAttrType, nextAttrValuecp ) ;

		    currAttributeLinep = currAttributeLinep -> nextAttrLinep ;

		    AllAttrValuesProcessed = TRUE ;
		}
	    }
	
	} while( AllAttrValuesProcessed == FALSE &&
		 newAttributeLinep != NULLALP ) ;


	if( newAttributeLinep == NULLALP )
	{
	    fprintf( stderr, "Memory allocation failure when creating a attributeLine node\n" );

	    (*finalNextAttrLineppp) = NULL ;
	}
	else
	    (*finalNextAttrLineppp) = &(currAttributeLinep -> nextAttrLinep) ;

	
	return( newAttributeLinep ) ; /* ptr may be NULLALP */

} /* create_attributeLine_node() */



struct rootedAtAttributes *
create_rootedAtAttrs_node( RDNLinep )
char  * RDNLinep ;
{
struct 	rootedAtAttributes * newRootedAtAttrsp = NULLRAA ;

	if( ( newRootedAtAttrsp = (struct rootedAtAttributes *)malloc( 
					(unsigned)sizeof( struct rootedAtAttributes )) )!= NULLRAA )
	{
totalRDNs++ ; /* statistics collection onlu */

	    newRootedAtAttrsp -> rdn = str2rdn( RDNLinep ) ;

	    if( newRootedAtAttrsp -> rdn == NULLRDN )
	    {
	    	fprintf( stderr, "FATAL ERROR : Could not parse %s\n", RDNLinep ) ;
		fprintf( stderr, "Will carry on looking for more problems but DISCARD output file\n" ) ;
	    }
	    
	    newRootedAtAttrsp -> rdnLine = strdup( RDNLinep ) ;

	    if( newRootedAtAttrsp -> rdnLine == NULLCP )
	    {
	    	fprintf( stderr, "FATAL ERROR : Memory allocation failure when duplicating %s\n", RDNLinep ) ;
		fprintf( stderr, "Will carry on looking for more problems but DISCARD output file\n" ) ;
	    }
#ifdef DEBUG2
fprintf( stderr, "Adding RDN %s", RDNLinep );
#endif
	    newRootedAtAttrsp -> attrsUsed = FALSE ;

	    newRootedAtAttrsp -> attrLinep = NULLALP ;

	    newRootedAtAttrsp -> nextRootedAtAttrsp = NULLRAA ;
	}

	if( newRootedAtAttrsp == NULLRAA )
	{
	    fprintf( stderr, "FATAL ERROR :Memory allocation failure when creating a rootedAtAttributes node\n" );
	    fprintf( stderr, "Will carry on looking for more problems but DISCARD output file\n" ) ;
	}

	return( newRootedAtAttrsp ) ; /* ptr may be NULLRAA */

} /* create_rootedAtAttrs_node() */



/*
 * Buffer the attribute lines for the current RDN
 */
 
int
bufferAttrs( DMfilep, lineBufferStructp, baseALpp)
FILE *	DMfilep ;
struct	lineBufferStruct * lineBufferStructp ;
struct	attributeLine ** baseALpp ;
{
struct	attributeLine ** currALpp = NULL, ** buffNewCurrALpp = NULL ;
int	AllAttrsRead = FALSE, ret_val = OK, res, atleastOneAttrRead = FALSE ;

	while( AllAttrsRead == FALSE && ret_val == OK )
	{
    	    if( (res = getLine( DMfilep, lineBufferStructp, ROOTEDATATTSORBLANKLINETYPE )) == OK )
    	    {
	    	if( determineLineType( lineBufferStructp -> lineBufferp ) == ATTSLINETYPE )
		{
                    atleastOneAttrRead = TRUE ;

		    /* Check that a value is present after the equals */
		    if( value_present( lineBufferStructp->lineBufferp ) == TRUE )
		    {		
		    	if( (*baseALpp) == NULLALP )
		      	{
		    	    (*baseALpp) = create_attributeLine_node( lineBufferStructp->lineBufferp, 
											&currALpp ) ;

		    	    if( !(*baseALpp) )
		    	    	ret_val = NOTOK ;		    
		    	}
		    	else
		    	{
		    	    (*currALpp) = create_attributeLine_node( lineBufferStructp -> lineBufferp, 
										&buffNewCurrALpp ) ;

		    	    if( !(*currALpp) )
		    	    	ret_val = NOTOK ;
		     	    else
		    	    	currALpp = buffNewCurrALpp ;
		    	}
		    }
		    
		    *(lineBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */
		}
		else AllAttrsRead = TRUE ;
	    }
	    else {

                if( res = NOTOKALLDATAREAD )
                {
		    AllAttrsRead = TRUE ;

		    if( atleastOneAttrRead == FALSE )
		    	ret_val = res ;
		}
	    	else
		    ret_val = res ; /* NOTOK - fatal error */
	    }
	    
	} /* while */

	return( ret_val ) ; /* OK, NOTOKALLDATAREAD or NOTOK */

} /* bufferAttrs() */



/*
 * RDNs must have a blank line preceding them.
 * A list of attributes will follow directly after.
 *
 * Note: "lineBuffer" may already have a line present.
 */
 
int
bufferRDNplusAttrs( DMfilep, lineBufferStructp, currRAApp )
FILE *	DMfilep ;
struct	lineBufferStruct * lineBufferStructp ;
struct  rootedAtAttributes ** currRAApp ;
{
int	getLineOK = OK, lineType = UNDEFLINETYPE ;
int	ret_val   = OK ;

	ret_val = getLineOK = getLine( DMfilep, lineBufferStructp, ROOTEDATATTSORBLANKLINETYPE ) ;

	if( getLineOK == OK  )
	{
	    lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;

	    if( lineType == BLANKLINETYPE )
	    {
	    	/* Discard any further sequential blank lines */
		while( getLineOK == OK && lineType == BLANKLINETYPE )
		{
		    *(lineBufferStructp -> lineBufferp ) = '\0' ; /* Discard blank line */

		    ret_val = getLineOK = getLine( DMfilep, lineBufferStructp, 
						ROOTEDATATTSORBLANKLINETYPE ) ;

		    lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;
		}

		if( getLineOK == OK  )
		{
	    	    lineType = determineLineType( lineBufferStructp -> lineBufferp ) ;

	    	    if( lineType == ATTSLINETYPE )
	    	    {
	    		/* Create a node to hold the retrieved RDN */
	    		(*currRAApp) = create_rootedAtAttrs_node( 
 					lineBufferStructp -> lineBufferp ) ;

	    		*(lineBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */

	    		if( !(*currRAApp) )
	    		{
			    ret_val == NOTOK ;
	    		}
	    		else
	    		{
			    /* Now load up the attributes to go with the RDN.. */
			    ret_val = bufferAttrs( DMfilep, lineBufferStructp, 
			    					&((*currRAApp) -> attrLinep)) ;
			}
		    }
		    else
		    {
			ret_val  =  NOTOKALLDATAREAD ; /* no RDN has been found */
	    	    }
		}
	    }
	    else
	    {
	    	ret_val  =  NOTOKALLDATAREAD ; /* no RDN has been found */
	    }
        }

	return( ret_val ) ; /* OK, NOTOKALLDATAREAD or NOTOK */

} /* bufferRDNplusAttrs() */


int
findRDNMatch( baseOldRAAp, baseNewRAAp, oldNodeNump, newNodeNump, differOldRAApp, differNewRAApp, LastAdvancedFile )
struct  rootedAtAttributes *  baseOldRAAp,    *  baseNewRAAp ;
int   * oldNodeNump, * newNodeNump ;
struct  rootedAtAttributes ** differOldRAApp, ** differNewRAApp ;  
int	LastAdvancedFile ;
{
int	matchFound = FALSE ;
struct  rootedAtAttributes * currOldRAAp = baseOldRAAp,
			   * currNewRAAp = baseNewRAAp ;

	if( !(baseOldRAAp) || !(baseNewRAAp) )
	    return( matchFound ) ;


	if( LastAdvancedFile == NEWFILE )
	{
	    *oldNodeNump = 0 ; *newNodeNump = 1 ;

	    /* Move the new file pointer to point to the latest RDN entry.. */

	    while( currNewRAAp && currNewRAAp -> nextRootedAtAttrsp )
	    {
		(*newNodeNump)++ ;

	    	currNewRAAp = currNewRAAp -> nextRootedAtAttrsp ;
	    }

	    
	    /* Traverse the OLDFILE linked list checking 
	       for a match with the NEWFILE RDN line.. */

	    while( currOldRAAp  && matchFound == FALSE )
	    {
		(*oldNodeNump)++ ;

		if( !rdn_cmp( currNewRAAp -> rdn, currOldRAAp -> rdn ))
{
		    matchFound = TRUE ;
#ifdef DEBUG3
fprintf( stderr, "RDN matched %s%s", currNewRAAp -> rdnLine , currOldRAAp -> rdnLine ) ;
#endif
}	    	else
	    	    currOldRAAp = currOldRAAp -> nextRootedAtAttrsp ;

	    } /* Old file while loop */
	}
	else /* OLDFILE was advanced */
	{
	    *newNodeNump = 0 ; *oldNodeNump = 1 ;

	    /* Move the old file pointer to point to the latest RDN entry.. */

	    while( currOldRAAp && currOldRAAp -> nextRootedAtAttrsp )
	    {
		(*oldNodeNump)++ ;
	    
	    	currOldRAAp = currOldRAAp -> nextRootedAtAttrsp ;
	    }

	    
	    /* Traverse the NEWFILE linked list checking 
	       for a match with the OLDFILE RDN line.. */

	    while( currNewRAAp && matchFound == FALSE )
	    {
		(*newNodeNump)++ ;
	
		if( !rdn_cmp( currNewRAAp -> rdn, currOldRAAp -> rdn ))
{
		    matchFound = TRUE ;
#ifdef DEBUG3
fprintf( stderr, "RDN matched %s%s", currNewRAAp -> rdnLine , currOldRAAp -> rdnLine ) ;
#endif
}
		else
	    	    currNewRAAp = currNewRAAp -> nextRootedAtAttrsp ;

	    } /* New file while loop */
	}


	if( matchFound == TRUE )
	{
	    (*differOldRAApp) = currOldRAAp ;
 	    (*differNewRAApp) = currNewRAAp ;
	}
	
	return( matchFound ) ; /* TRUE or FALSE */
	
} /* findRDNMatch() */


/*
 * Compare the two values whilst ignoring all 
 * white space characters..
 */
 
int
nonWspceStrsMatch( string1, string2 )
char *	string1, * string2 ;
{
int	matched = TRUE ;
char *  str1cp = string1, * str2cp = string2 ;
int	ptrIncremented, AnyMoreWspecChars = FALSE ;

	if( string1 == NULLCP || string2 == NULLCP )
	    return( FALSE ) ;


	while( *str1cp && *str2cp && matched == TRUE )
	{
	    ptrIncremented = FALSE ;
	    
	    if( isspace( *str1cp ) )
	    {
	    	str1cp++ ;
		ptrIncremented = TRUE ;
	    }
	    
	    if( isspace( *str2cp ) )
	    {
	    	str2cp++ ;
		ptrIncremented = TRUE ;
	    }


	    if( ptrIncremented == FALSE )
	    {
	    	if( *str1cp != *str2cp )
	    	    matched = FALSE ;

	    	else {
	    	    str1cp++, str2cp++ ;
	    	}
	    }
	}

	if( matched == TRUE )
	{
	    /* Check to see if there are any unread non-white
	       space characters in either string..            */

	    while( *str1cp && AnyMoreWspecChars == FALSE )
	    {
	        if( !isspace( *str1cp++ ) )
	   	    AnyMoreWspecChars = TRUE ;
	    }


	    while( *str2cp && AnyMoreWspecChars == FALSE )
	    {
	    	if( !isspace( *str2cp++ ) )
	    	    AnyMoreWspecChars = TRUE ;
	    }

	    if( AnyMoreWspecChars == TRUE )
	    	matched = FALSE ;
	}


	return( matched ) ;
}


int
attrValueMatch( AL1p, AL2p )
struct attributeLine * AL1p, * AL2p ;
{
int	matched = FALSE ;
AttributeValue av1, av2 ;
char *  str1cp, * str2cp ;

	if( AL1p == NULLALP || AL2p == NULLALP )
	    return( matched ) ;


	if( AL1p -> at->oa_syntax == 0 || AL2p -> at->oa_syntax == 0 )
	{
	    /* Unknown syntax e.g. {ASN}.. */

	    if( (str1cp = strchr( AL1p->attrLine, '=' )) != NULLCP &&
		(str2cp = strchr( AL2p->attrLine, '=' )) != NULLCP   )
	    	matched = nonWspceStrsMatch( str1cp+1, str2cp+1 ) ; /* ..Skipping '='s */
	    else
	    	matched = FALSE ;
	}
	else
	{
	    /* Known syntax.. */

	    av1 = AL1p -> av , av2 = AL2p -> av ;

	    if( av1 != NULLAttrV && av2 != NULLAttrV )
	    {
	   	if( !AttrV_cmp( av1, av2 ) )
		    matched = TRUE ;
	    }
	}
	
	return( matched ) ;

} /*attrValueMatch() */


int
attrTypeMatch( at1, at2 )
AttributeType at1, at2 ;
{
int	matched = FALSE ;

	if( at1 != NULLAttrT && at2 != NULLAttrT )
	{
	    if( !AttrT_cmp( at1, at2 ) )
		matched = TRUE ;
	}

	return( matched ) ; /* TRUE or FALSE */

} /* attrTypeMatch() */


/*
 * Is the attribute type used within the RDN ?
 */
 
int
attrTypeInRDN( at, rdn )
AttributeType	at  ;
RDN		rdn ;
{
RDN	currRdncomp = rdn ;
int	loopExit    = FALSE ;

	while( currRdncomp != NULLRDN && loopExit == FALSE )
	{
	    if( AttrT_cmp( currRdncomp->rdn_at, at ) == 0 )
		loopExit = TRUE ;
	    else
	    	currRdncomp = currRdncomp -> rdn_next ;
	}

	if( loopExit == TRUE )
	    return( TRUE ) ;
	else
	    return( FALSE ) ;
}


/* 
 * Find any sequences of attributes of the same type
 * within the specified attributeLine list.
 */

void
findAttrTypeSequences( baseAttrLinep )
struct	attributeLine * baseAttrLinep ;
{
struct	attributeLine * currAttrLinep = baseAttrLinep, * checkALp = NULLALP ;
int	attrTypeMatchCount ;
AttributeType 	seqAt = NULLAttrT ;

	while( currAttrLinep != NULLALP )
	{
	    if( currAttrLinep -> nthAttrOfThisType == 0 )
	    {
	    	seqAt = currAttrLinep -> at ;

	    	attrTypeMatchCount = 1 ;

	    	currAttrLinep -> nthAttrOfThisType = attrTypeMatchCount++ ;

		/* Run through the remaining attributes checking
		   for those with the same attribute type */
	    	checkALp = currAttrLinep -> nextAttrLinep ;

		while( checkALp != NULLALP )
	    	{
	    	    if( checkALp -> nthAttrOfThisType == 0 &&
		        attrTypeMatch( checkALp->at, seqAt ) == TRUE )
		    {
		    	checkALp -> nthAttrOfThisType = attrTypeMatchCount++ ;
		    }

	    	    checkALp = checkALp -> nextAttrLinep ;
		}
	    }
	    
	    currAttrLinep = currAttrLinep -> nextAttrLinep ;
	}

	return ;
}


/*
 * Search the attributeLine linked list for
 * any attribute having the specified type.
 */
 
int
findAttrType( baseAttrLinep, findAt )
struct	attributeLine * baseAttrLinep ; 
AttributeType findAt ;
{
struct	attributeLine * currAttrLinep = baseAttrLinep ;
int	matchFound = FALSE ;

	while( currAttrLinep && matchFound == FALSE )
	{
	    if( attrTypeMatch( currAttrLinep -> at, findAt ) == TRUE )
	    	matchFound = TRUE ;
	    else
		currAttrLinep = currAttrLinep -> nextAttrLinep ;
	}

	return( matchFound ); /* TRUE or FALSE */
}


/*
 * Compare two attributeLine linked lists to determine
 * which attributes appear in both.
 *
 * For those attribute values that appear solely in the new source
 * determine whether an attribute or an attribute value addition should
 * be indicated.
 *
 * For those attribute values that appear solely in the old source
 * indicate an attribute value deletion.
 *
 * Note: the procedure must take into account the attribute type of
 * the RDN when determining whether attributes are being added or
 * deleted for the first time.
 */

void
differAttributesFromRAAnode( differOldRAAp, differNewRAAp )
struct  rootedAtAttributes * differOldRAAp, * differNewRAAp ;
{
struct	attributeLine * baseOldAttrLinep  = (differOldRAAp -> attrLinep) ;
struct	attributeLine * baseNewAttrLinep  = (differNewRAAp -> attrLinep) ;
struct	attributeLine * currOldAttrLinep , * currNewAttrLinep ;
int	RDNdisplayed = FALSE ;

	/* Set up the AttibuteValue and AttributeType variables.. */
	setUpAttrTypeValue( baseOldAttrLinep ) ;
	setUpAttrTypeValue( baseNewAttrLinep ) ;


	currOldAttrLinep = baseOldAttrLinep ;
	currNewAttrLinep = baseNewAttrLinep ;

	/* Check for each attribute from the Old file within the New files attributes */

	while( currOldAttrLinep )
	{
	    currNewAttrLinep = baseNewAttrLinep ;	

	    while( currNewAttrLinep && currOldAttrLinep -> attrUsed == FALSE )
	    {
	    	if( currNewAttrLinep -> attrUsed == FALSE )
		{
		    /* Check that the attribute types match.. */
		    if( attrTypeMatch( currOldAttrLinep -> at, 
		    				currNewAttrLinep -> at ) == TRUE )
		    {
	    	    	if( attrValueMatch
				( currOldAttrLinep, currNewAttrLinep ) == TRUE )
		    	{
		    	    /* No change as attribute listed under both sources */

			    currOldAttrLinep -> attrUsed = TRUE ;
			    currNewAttrLinep -> attrUsed = TRUE ;
			}
		    }
		}

		currNewAttrLinep = currNewAttrLinep -> nextAttrLinep ;
	    }

	    currOldAttrLinep = currOldAttrLinep -> nextAttrLinep ;
	}

	
	/* Now display those attributes and attribute values that are to be added.. */

	/* Find any sequences of unused attributes of the same type within
	   the attributeLine list and set up the nthAttrOfThisType field */
	
	findAttrTypeSequences( baseNewAttrLinep ) ;

	currNewAttrLinep = baseNewAttrLinep ;

	while( currNewAttrLinep )
	{
	    if( currNewAttrLinep -> attrUsed == FALSE )
	    {
		if( RDNdisplayed == FALSE )
		{
		    printf( "\n%s", differOldRAAp -> rdnLine ) ;
		    RDNdisplayed = TRUE ;
		}

	    	/* Determine whether this is an attribute or
		   an attribute value addition..            */

	    	if( currNewAttrLinep -> nthAttrOfThisType > 1 ||
		    attrTypeInRDN( currNewAttrLinep -> at, differNewRAAp -> rdn ) == TRUE || /* an RDN component is of this type */
		    findAttrType( baseOldAttrLinep, currNewAttrLinep -> at ) == TRUE )

	    	    printf( "%c%s", ADDVALUECHAR, currNewAttrLinep -> attrLine ) ;
		else
	    	    printf( "%c%s", ADDATTRCHAR, currNewAttrLinep -> attrLine ) ;

	    	currNewAttrLinep -> attrUsed = TRUE ; /* For completeness */
	    }

	    currNewAttrLinep = currNewAttrLinep -> nextAttrLinep ;	    	
	}


	/*
	 * Now display those attributes values that are to be deleted..
	 */
	 
	currOldAttrLinep = baseOldAttrLinep ;

	while( currOldAttrLinep )
	{
	    if( currOldAttrLinep -> attrUsed == FALSE )
	    {
		if( RDNdisplayed == FALSE )
		{
		    printf( "\n%s", differOldRAAp -> rdnLine ) ;
		    RDNdisplayed = TRUE ;
		}

		printf( "%c%s", DELETEVALUECHAR, currOldAttrLinep -> attrLine ) ;

	    	currOldAttrLinep -> attrUsed = TRUE ; /* For completeness */
	    }
	    
	    currOldAttrLinep = currOldAttrLinep -> nextAttrLinep ;	    	
	}

	/* Clear the AttibuteValue and AttributeType storage.. */
	clearDownAttrTypeValue( baseOldAttrLinep ) ;
	clearDownAttrTypeValue( baseNewAttrLinep ) ;

	
	return ;
}


/*
 * Determine whether the individual attribute values are
 * being added for the first time, taking account of
 * the attribute type of the RDN.
 */

void
outputAttrsFromAttrLineList( currRAAp )
struct  rootedAtAttributes * currRAAp ;
{
struct 	attributeLine * baseALp = currRAAp -> attrLinep, * currALp ;

	/* Set up the AttibuteValue and AttributeType variables.. */
	setUpAttrTypeValue( baseALp ) ;

	findAttrTypeSequences( baseALp ) ;

	currALp = baseALp ;

	while( currALp )
	{
	    if( currALp -> nthAttrOfThisType > 1 ||
	        attrTypeInRDN( currALp->at, currRAAp -> rdn ) == TRUE ) /* an RDN component is of this type */
		printf( "%c%s", ADDVALUECHAR, currALp -> attrLine ) ;
	    else
		printf( "%c%s", ADDATTRCHAR,  currALp -> attrLine ) ;

	    currALp = currALp -> nextAttrLinep ;
	}

	clearDownAttrTypeValue( baseALp ) ;

	return ;

} /* outputAttrsFromAttrLineList() */


void
outputAddedRDNandAttrsFromRAAlist( baseRAAp, numberOfAddedNodes )
struct  rootedAtAttributes * baseRAAp ;
int	numberOfAddedNodes ; /* may be set to ALLNODES */
{
struct  rootedAtAttributes * currRAAp = baseRAAp ;
int	nodeNum = 0 ;

	while( currRAAp && 
		( nodeNum < numberOfAddedNodes || numberOfAddedNodes == ALLNODES ))
	{
	    printf( "\n%c%s", ADDRDNCHAR, currRAAp -> rdnLine ) ;
	    
	    outputAttrsFromAttrLineList( currRAAp ) ;
	    
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;

	    nodeNum++ ;
	}

} /* outputAddedRDNandAttrsFromRAAlist() */



void
outputDeletedRDNsFromRAAlist( baseRAAp, numberOfDeleteNodes, deleteAttributeNotEntries )
struct  rootedAtAttributes * baseRAAp ;
int	numberOfDeleteNodes ; /* may be set to ALLNODES */
int	deleteAttributeNotEntries ;
{
struct  rootedAtAttributes * currRAAp = baseRAAp ;
struct	attributeLine * currALp = NULLALP ;
int	nodeNum = 0 ;
static	AttributeType objclsat = NULLAttrT ;

	if( objclsat == NULLAttrT )
	    objclsat =  str2AttrT( "objectClass" ) ;

	while( currRAAp && 
		( nodeNum < numberOfDeleteNodes || numberOfDeleteNodes == ALLNODES ))
	{
	    if( deleteAttributeNotEntries == DELENTRIESNOTATTRS )
	    {
	    	printf( "\n%c%s", DELETERDNCHAR, currRAAp -> rdnLine ) ;
	    }
	    else
	    {
	    	printf( "\n%s", currRAAp -> rdnLine ) ;

		setUpAttrTypeValue( currRAAp -> attrLinep ) ;

		for( currALp = currRAAp -> attrLinep ; currALp != NULLALP ; currALp = currALp -> nextAttrLinep )
		{
		    if( deleteAttributeNotEntries == DELATTRSNOTENTRIES ||
		    	(currALp->at != NULLAttrT &&  AttrT_cmp( currALp->at, objclsat ) != 0) ) /* ...DELATTRSNOTOBJCLSORENTS */
		    {
		    	printf( "%c%s", DELETEVALUECHAR, currALp -> attrLine ) ;
		    }
		}

		clearDownAttrTypeValue( currRAAp -> attrLinep ) ;
	    }
	    
	    currRAAp = currRAAp -> nextRootedAtAttrsp ;

	    nodeNum++ ;
	}

} /* outputDeletedRDNsFromRAAlist() */




int
differAttributes( oldDMfilep, oldLineBufferStructp, newDMfilep, newLineBufferStructp, deleteAttributeNotEntries )
FILE  * oldDMfilep, * newDMfilep ;
struct	lineBufferStruct * oldLineBufferStructp, * newLineBufferStructp ;
int	deleteAttributeNotEntries ;
{
int	ret_val = OK, RDNfound,
	oldNodeNum, newNodeNum, getOldRDNblockOK = OK, getNewRDNblockOK = OK ;
int	LastAdvancedFile = OLDFILE, NextAdvancedFile = NEWFILE,
	AllRDNsLoopExit = FALSE ;
struct  rootedAtAttributes *  baseOldRAAp   = NULLRAA, 
			   *  baseNewRAAp   = NULLRAA,
			   ** currOldRAApp  = &baseOldRAAp,
			   ** currNewRAApp  = &baseNewRAAp,
 			   *  differOldRAAp = NULLRAA,
			   *  differNewRAAp = NULLRAA ;
			   
	AllRDNsLoopExit = FALSE ;

	while( AllRDNsLoopExit == FALSE && ret_val == OK )
	{
	    RDNfound = FALSE ;

	    while( AllRDNsLoopExit == FALSE && ret_val == OK && RDNfound == FALSE )
	    {
	    	/* Retrieve the next RDN and attributes block *
 	    	 * from either the Old or the New files       */
#ifdef DEBUG
fprintf( stderr, "Next ad=%d\n", NextAdvancedFile );
#endif	       
	    	if( NextAdvancedFile == OLDFILE )
	    	{
	    	    if( (getOldRDNblockOK = 
		    	 bufferRDNplusAttrs( oldDMfilep, oldLineBufferStructp, currOldRAApp )) == OK )
	    	    {
			currOldRAApp = &((*currOldRAApp)-> nextRootedAtAttrsp) ;
	    	    }
		    
		    if( getOldRDNblockOK == NOTOK )
		    	ret_val = NOTOK ;

		    LastAdvancedFile = OLDFILE ;
	    	}
	    	else
	    	{   /* Advance new file */

	    	    if( (getNewRDNblockOK =
		    	 bufferRDNplusAttrs( newDMfilep, newLineBufferStructp, currNewRAApp )) == OK )
	    	    {
			currNewRAApp = &((*currNewRAApp)-> nextRootedAtAttrsp) ;
	    	    }
		    
		    if( getNewRDNblockOK == NOTOK )
		    	ret_val = NOTOK ;

		    LastAdvancedFile = NEWFILE ;
	    	}


	    	/* See if an RDN exists in both the new and old lists */
	    	if( ret_val == OK )
	    	{
		    if( findRDNMatch( baseOldRAAp, baseNewRAAp, &oldNodeNum, &newNodeNum,
				&differOldRAAp, &differNewRAAp, LastAdvancedFile ) == TRUE )
		    {

#ifdef DEBUG
fprintf( stderr, "RDN Matched %d %d\n", oldNodeNum, newNodeNum );
#endif

			differAttributesFromRAAnode( differOldRAAp, differNewRAAp ) ;

#ifdef DEBUG
fprintf( stderr, "finished differing attrs \n" ) ;
#endif
		    	RDNfound = TRUE ;
		    }
	    	}

		
		if( getOldRDNblockOK == NOTOK || getNewRDNblockOK == NOTOK ||
		    (getOldRDNblockOK == NOTOKALLDATAREAD && getNewRDNblockOK == NOTOKALLDATAREAD) )
		    AllRDNsLoopExit = TRUE ;

		if( LastAdvancedFile == OLDFILE && getNewRDNblockOK == OK )
		    NextAdvancedFile =  NEWFILE ;

		else if( LastAdvancedFile == NEWFILE && getOldRDNblockOK == OK )
		    NextAdvancedFile =  OLDFILE ;

	    } /* RDN attributes while loop */

#ifdef DEBUG
fprintf( stderr, "[] %d %d %d\n", AllRDNsLoopExit, ret_val, RDNfound ) ;
#endif
	    if( RDNfound == TRUE )
	    {
		/* We must unlink the matched nodes */

		currOldRAApp = freeRootAtAttrs_node( &baseOldRAAp, oldNodeNum ) ;
		currNewRAApp = freeRootAtAttrs_node( &baseNewRAAp, newNodeNum ) ;
	    }
	    
	} /* rootedAt RDNs while */

#ifdef DEBUG
fprintf( stderr, "RET VAL %d AllRDNsLoopExit %d RDNfound %d\n", ret_val , AllRDNsLoopExit, RDNfound) ;
#endif
	if( ret_val != NOTOK )
	{
	    /*
	     * Check to see if either files have RDN/attributes
	     * remaining within the current "rootedAt" block that have not been read.
	     */
	    if( getOldRDNblockOK != NOTOK )
	    {
	    	if( baseOldRAAp == NULLRAA )
		    currOldRAApp = &baseOldRAAp;

	    	do {

	            if( (getOldRDNblockOK = 
		     	bufferRDNplusAttrs( oldDMfilep, oldLineBufferStructp, currOldRAApp )) == OK )
	    	    {
		    	currOldRAApp = &((*currOldRAApp)-> nextRootedAtAttrsp) ;
		    }
		
	    	} while( getOldRDNblockOK == OK ) ;

	    	if( getOldRDNblockOK == NOTOK ) ret_val = NOTOK ;
	    }
	    
	    if( getNewRDNblockOK != NOTOK )
	    {
	    	if( baseNewRAAp == NULLRAA )
		    currNewRAApp = &baseNewRAAp;
	    
	    	do {
	    	    if( (getNewRDNblockOK = 
		    	 bufferRDNplusAttrs( newDMfilep, newLineBufferStructp, currNewRAApp )) == OK )
	    	    {
		    	currNewRAApp = &((*currNewRAApp)-> nextRootedAtAttrsp) ;
		    }
		    
	    	} while( getNewRDNblockOK == OK ) ;

	    	if( getNewRDNblockOK == NOTOK ) ret_val = NOTOK ;
	    }


	    
	    /*
	     * Process any unused data within the RAA linked lists
	     * (whether from the block immediately above or from the
	     *  RDN matching while loops)
	     */

	    /* Process any deleted RDNs */
	    if( baseOldRAAp )
	    	outputDeletedRDNsFromRAAlist( baseOldRAAp, ALLNODES, deleteAttributeNotEntries ) ;	

	    /* Process any added RDNs and their attributes */
	    if( baseNewRAAp )
		outputAddedRDNandAttrsFromRAAlist( baseNewRAAp, ALLNODES ) ;
    
	    /*
	     * Finally return any remaining storage
	     */
	    freeRootAtAttrs_list( &baseOldRAAp ) ;
	    freeRootAtAttrs_list( &baseNewRAAp ) ;
	}


	return( ret_val ) ;

} /* differAttributes() */



/*
 * Note: uses buffers the current line buffer contents since
 * they do not apply after the fseek(). The buffer contents
 * must be returned when the file positions are themselves
 * reinstated.
 */

int
differAttributesFromRADnode( oldDMfilep, oldCurrRADp, oldLineBufferStructp, 
				newDMfilep, newCurrRADp, newLineBufferStructp, deleteAttributeNotEntries )
FILE *	oldDMfilep, * newDMfilep ;
struct	rootedAtDetails * oldCurrRADp, * newCurrRADp ;
struct	lineBufferStruct * oldLineBufferStructp, * newLineBufferStructp ;
int	deleteAttributeNotEntries ;
{
char *	savedOldLineBufferp = NULLCP, * savedNewLineBufferp = NULLCP ;
long	oldOrigFileOffset = ftell( oldDMfilep ),
	newOrigFileOffset = ftell( newDMfilep ) ;
int	ret_val = OK ;

	/* Save both line buffers */
	if( (savedOldLineBufferp = strdup( oldLineBufferStructp -> lineBufferp )) == NULLCP ||
	    (savedNewLineBufferp = strdup( newLineBufferStructp -> lineBufferp )) == NULLCP )
	{
	    fprintf( stderr, "Memory allocation failure when saving the line buffers\n" );

	    return( NOTOK ) ;
	}


	/* Clear down the line buffers now that the file pointers are to be moved.. */
	
	*(oldLineBufferStructp -> lineBufferp) =
		*(newLineBufferStructp -> lineBufferp) = '\0' ;
		

	/* Adjust both file pointers to the required rootedAt offset.. */
	if( fseek( oldDMfilep, oldCurrRADp -> fileOffset, 0 ) ||
	    fseek( newDMfilep, newCurrRADp -> fileOffset, 0 )   )
	    return( NOTOK ) ;


	/* Output the rootedAt line.. */
	printf( "\n\n%s\n", newCurrRADp -> rootedAtLine ) ;

	if( differAttributes( oldDMfilep, oldLineBufferStructp, 
				newDMfilep, newLineBufferStructp, deleteAttributeNotEntries ) == NOTOK )
	    ret_val = NOTOK ;

	
	/* Return the file pointers to their original positions */
	if( fseek( oldDMfilep, oldOrigFileOffset, 0 ) ||
	    fseek( newDMfilep, newOrigFileOffset, 0 )   )
	    ret_val = NOTOK ;

	/* Reinstate the line buffers */
	if( savedOldLineBufferp != NULLCP )
	{
	    strcpy( oldLineBufferStructp -> lineBufferp, savedOldLineBufferp ) ;
	    free( savedOldLineBufferp ) ;
	}
	
	if( savedNewLineBufferp != NULLCP )
	{
	    strcpy( newLineBufferStructp -> lineBufferp, savedNewLineBufferp ) ;
	    free( savedNewLineBufferp ) ;
	}

	return( ret_val ) ;
}


/************************************************/




void
freeRootAtDetails_list( baseRootedAtDetailsp )
struct rootedAtDetails * baseRootedAtDetailsp ;
{
struct rootedAtDetails * currRootedAtDetailsp = baseRootedAtDetailsp,
		       * buffRootedAtDetailsp = NULLRAD ;

	while( currRootedAtDetailsp != NULLRAD )
	{
	    /* Free the character line storage.. */
	    (void)free((char*)(currRootedAtDetailsp -> rootedAtLine) ) ;

	    buffRootedAtDetailsp = currRootedAtDetailsp ;
	    currRootedAtDetailsp = currRootedAtDetailsp -> nextRootedAtDetailsp ;

	    /* Free the node storage.. */
	    (void)free((char*)( buffRootedAtDetailsp) ) ;
	}

	return ;
}



struct rootedAtDetails *
freeRootAtDetails_node( baseRADpp, nodeNumToBeUnlinked )
struct rootedAtDetails ** baseRADpp ;
{
struct rootedAtDetails * currRADp = * baseRADpp,
		       * buffRADp = NULLRAD,
		       * prevRADp = NULLRAD,
		       * lastRADp = NULLRAD ;
int	currNodeNum = 1 ;


	/* Note: Nodes number 1,2, .. */

	/* Find the node that is to be unlinked.. */
	while( currRADp != NULLRAD && currNodeNum < nodeNumToBeUnlinked )
	{
	    currNodeNum++ ;

	    prevRADp = currRADp ;
	    currRADp = currRADp -> nextRootedAtDetailsp ;
	}
#ifdef DEBUG1
fprintf( stderr, "Freeing node num %d\n%s",
	currNodeNum, currRADp -> rootedAtLine );
#endif
	if( nodeNumToBeUnlinked == currNodeNum )
	{
	    /* Free the character line storage.. */
	    (void)free((char*)(currRADp -> rootedAtLine) ) ;

	    buffRADp = currRADp ;
	    currRADp = currRADp -> nextRootedAtDetailsp ;

	    /* Free the node storage.. */
	    (void)free((char*)( buffRADp) ) ;

	    /* Link around the freed node */
	    if( prevRADp )
	    	prevRADp -> nextRootedAtDetailsp = currRADp ;

	    if( currNodeNum == 1 )
	    {
	    	/* .. have freed the base node */
	    
  	    	*baseRADpp = currRADp ;
	    }

	    
	    /* Determine the last node.. */
	    if( currRADp == NULLRAD )
	    	lastRADp =  prevRADp ; /* NULLRAD if sole node deleted */
	    else
	    {
		/* Find the last node */
		while( currRADp != NULLRAD )
		{
	    	    buffRADp = currRADp ;
	    	    currRADp = currRADp -> nextRootedAtDetailsp ;
		}

		lastRADp = buffRADp ;
	    }
	}
	else
	{
	    /* Error condition */
	    freeRootAtDetails_list( *baseRADpp ) ;
	}


	/* Return a pointer to the last node in the list.. */
	if( lastRADp )
	    return( lastRADp );
	else
	    return( NULLRAD ) ;
}


/*
 * Create a node containing the new offset and line details to the rootedAt linked list.
 */

struct rootedAtDetails *
create_rootedAtDetails_node( fileOffset, rootedAtLinep )
long	fileOffset ;
char  * rootedAtLinep ;
{
struct 	rootedAtDetails * newRootedAtDetailsp ;

	if( ( newRootedAtDetailsp = (struct rootedAtDetails *)malloc( 
					(unsigned)sizeof( struct rootedAtDetails )) )!= NULLRAD )
	{
	    newRootedAtDetailsp -> fileOffset = fileOffset ;
	    newRootedAtDetailsp -> rootedAtLine = strdup( rootedAtLinep ) ;

#ifdef DEBUG
fprintf( stderr, "create rootedAtDetails\n %s", rootedAtLinep );
#endif
	    newRootedAtDetailsp -> nextRootedAtDetailsp = NULLRAD ;
	}

	if( newRootedAtDetailsp == NULLRAD )
	    fprintf( stderr, "Memory allocation failure when creating a rootedAt node\n" );

	return( newRootedAtDetailsp ) ;
}



/* NOT CURRENTLY USED */
long
rootedAtDetailsFileOffset( baseRADp, nodeNum )
struct 	rootedAtDetails * baseRADp ;
int	nodeNum ;
{
long	fileOffset = 0 ;
struct	rootedAtDetails * currRADp = baseRADp ;
int	currNodeNum = 0 ;

	while( ++currNodeNum <  nodeNum &&
	       currRADp      != NULLRAD   )
	{
	    currRADp = currRADp -> nextRootedAtDetailsp ;
	    currNodeNum++ ;
	}

	/* If the required node exists then buffer the file offset field..*/
	if( currNodeNum == nodeNum &&
	    currRADp    != NULLRAD      )
	{
	    fileOffset  =  currRADp -> fileOffset ;
	}
	
	return( fileOffset ) ;
}


/*
 * Note: We know that the "NumberOfDeleteNodes is atleast 1 less
 * than the total number of nodes within the linked list.
 */
 
int
deleteEntriesFromRADlist( DMfilep, baseRADp, lineBufferStructp, NumberOfDeleteNodes, deleteAttributeNotEntries )
FILE  * DMfilep ;
struct	rootedAtDetails * baseRADp ;
struct	lineBufferStruct * lineBufferStructp ;
int	NumberOfDeleteNodes, deleteAttributeNotEntries ;
{
struct	rootedAtDetails * currRADp = baseRADp ;
struct  rootedAtAttributes * baseRAAp = NULLRAA, ** currRAApp = &baseRAAp ;
int	currNodeNum = 0 ;
long	origFileOffset = ftell( DMfilep ),
	currUpperBoundFileOffset ;
int	ret_val = OK, getRDNATTRSOK = OK ;


	while( (currNodeNum < NumberOfDeleteNodes || NumberOfDeleteNodes == ALLNODES ) &&
		currRADp && getRDNATTRSOK == OK )
	{
	    *(lineBufferStructp -> lineBufferp ) = '\0' ;

	    /* Set the file offset to the position of the current rootedAt
	       line which designates an entry to be deleted.. */

	    fseek( DMfilep, currRADp -> fileOffset, 0 );
	    
	    if( currRADp -> nextRootedAtDetailsp )
	    	currUpperBoundFileOffset = currRADp -> nextRootedAtDetailsp -> fileOffset ;
	    else
		currUpperBoundFileOffset = BOUNDEDBYEOF ;
		

	    printf( "\n%s\n", currRADp -> rootedAtLine ) ;


	    while( getRDNATTRSOK == OK &&
	    	   (ftell( DMfilep ) < currUpperBoundFileOffset || 
		   	currUpperBoundFileOffset == BOUNDEDBYEOF  ) )
	    {
	    	getRDNATTRSOK = bufferRDNplusAttrs( DMfilep, lineBufferStructp, 
									currRAApp ) ;

	    	if( getRDNATTRSOK == OK )
		    currRAApp = &((*currRAApp) -> nextRootedAtAttrsp ) ;
	    }

	    if( getRDNATTRSOK != NOTOK )
	    	getRDNATTRSOK = OK ;

	    /* .. loop should exit with a getRDNATTRSOK == NOTOKALLDATAREAD
	       to indicate that the current "rootedAt" block has no more RDNs */

	    if( getRDNATTRSOK == OK )
	    	outputDeletedRDNsFromRAAlist( baseRAAp, ALLNODES, deleteAttributeNotEntries ) ;
	    


	    freeRootAtAttrs_list( &baseRAAp ) ;
	    currRAApp = & baseRAAp ; /* Reset the buffer pointer */

	    
	    currRADp = currRADp -> nextRootedAtDetailsp ;
	    currNodeNum++ ;
	}


	if( getRDNATTRSOK == NOTOK )
 	{
	    fprintf( stderr, "Deleted Entry output failed\n" ) ;
	    ret_val = NOTOK ;
	}


	/* Return the file pointer to its original position */
	(void)fseek( DMfilep, origFileOffset, 0 );

	return ret_val ;
}




/*
 * Note: We know that the "NumberOfAddNodes is atleast 1 less
 * than the total number of nodes within the linked list.
 */
 
int
addEntriesFromRADlist( DMfilep, baseRADp, lineBufferStructp, NumberOfAddNodes )
FILE  * DMfilep ;
struct	rootedAtDetails * baseRADp ;
struct	lineBufferStruct * lineBufferStructp ;
int	NumberOfAddNodes ;
{
struct	rootedAtDetails * currRADp    = baseRADp ;
struct  rootedAtAttributes * baseRAAp = NULLRAA, ** currRAApp = &baseRAAp ;
int	currNodeNum = 0 ;
long	origFileOffset = ftell( DMfilep ),
	currUpperBoundFileOffset ;
int	ret_val = OK, getRDNATTRSOK = OK ;

	while( (currNodeNum < NumberOfAddNodes || NumberOfAddNodes == ALLNODES ) &&
	       currRADp && getRDNATTRSOK == OK )
	{
	    *(lineBufferStructp -> lineBufferp) = '\0' ;

	    /* Set the file offset to the position of the current rootedAt
	       line which designates an entry to be added.. */

	    (void)fseek( DMfilep, currRADp -> fileOffset, 0 ) ;

	    if( currRADp -> nextRootedAtDetailsp )
	    	currUpperBoundFileOffset = currRADp -> nextRootedAtDetailsp -> fileOffset ;
	    else
		currUpperBoundFileOffset = BOUNDEDBYEOF ;


	    printf( "\n%s\n", currRADp -> rootedAtLine ) ;


	    while( getRDNATTRSOK == OK &&
	    	   (ftell( DMfilep ) < currUpperBoundFileOffset || 
		   	currUpperBoundFileOffset == BOUNDEDBYEOF  ) )
	    {
	    	getRDNATTRSOK = bufferRDNplusAttrs( DMfilep, lineBufferStructp, currRAApp ) ;

	    	if( getRDNATTRSOK == OK )
		    currRAApp = &((*currRAApp) -> nextRootedAtAttrsp ) ;
	    }


	    if( getRDNATTRSOK != NOTOK )
	    	getRDNATTRSOK = OK ;
		
	    /* .. loop should exit with a getRDNATTRSOK == NOTOKALLDATAREAD
	       to indicate that the current "rootedAt" block has no more RDNs */


	    if( getRDNATTRSOK == OK )
	    	outputAddedRDNandAttrsFromRAAlist( baseRAAp, ALLNODES ) ;

	    freeRootAtAttrs_list( &baseRAAp ) ;
	    currRAApp = & baseRAAp ; /* Reset the buffer pointer */
	    

	    currRADp = currRADp -> nextRootedAtDetailsp ;
	    currNodeNum++ ;
	}
	

	if( getRDNATTRSOK == NOTOK )
 	{
	    fprintf( stderr, "Add Entry output failed\n" ) ;
	    ret_val = NOTOK ;
	}


	/* Return the file pointer to its original position */
	(void)fseek( DMfilep, origFileOffset, 0 );

	return( ret_val ); /* OK or NOTOK */
}



int
findRootedAtMatch( baseOldRADp, baseNewRADp, oldNodeNump, newNodeNump, 
			differOldRADpp, differNewRADpp, LastAdvancedFile )
struct 	rootedAtDetails  * baseOldRADp, * baseNewRADp ;
int   * oldNodeNump,      * newNodeNump ;
struct 	rootedAtDetails ** differOldRADpp, ** differNewRADpp ;
int	LastAdvancedFile ;
{
int	matchFound = FALSE ;
struct 	rootedAtDetails * currOldRADp = baseOldRADp,
			* currNewRADp = baseNewRADp ;


	/* Nodes shall be numbered 1,2.. */
	

	if( LastAdvancedFile == NEWFILE )
	{
	    *oldNodeNump = 0 ; *newNodeNump = 1 ;

	    /* Move the new file pointer to point to the latest entry.. */

	    while( currNewRADp && currNewRADp -> nextRootedAtDetailsp )
	    {
		(*newNodeNump)++ ;

	    	currNewRADp = currNewRADp -> nextRootedAtDetailsp ;
	    }

	    
	    /* Traverse the OLDFILE linked list checking 
	       for a match with the NEWFILE rootedAt line.. */

	    while( currOldRADp != NULLRAD && matchFound == FALSE )
	    {
		(*oldNodeNump)++ ;
	
		if( !strcasecmp( currNewRADp -> rootedAtLine, currOldRADp -> rootedAtLine ) )
		    matchFound = TRUE ;
	    	else
	    	    currOldRADp = currOldRADp -> nextRootedAtDetailsp ;

	    } /* Old file while loop */
	}
	else if( LastAdvancedFile == OLDFILE )
	{
	    *newNodeNump = 0 ; *oldNodeNump = 1 ;

	    /* Move the old file pointer to point to the latest entry.. */

	    while( currOldRADp && currOldRADp -> nextRootedAtDetailsp )
	    {
		(*oldNodeNump)++ ;
	    
	    	currOldRADp = currOldRADp -> nextRootedAtDetailsp ;
	    }

	    
	    /* Traverse the NEWFILE linked list checking 
	       for a match with the OLDFILE rootedAt line.. */

	    while( currNewRADp != NULLRAD && matchFound == FALSE )
	    {

		(*newNodeNump)++ ;
	
		if( !strcasecmp( currNewRADp -> rootedAtLine, currOldRADp -> rootedAtLine ) )
		    matchFound = TRUE ;
	    	else
	    	    currNewRADp = currNewRADp -> nextRootedAtDetailsp ;

	    } /* New file while loop */
	}
	else /* BOTHFILES were advanced */
	{
	    *newNodeNump = 0 ; *oldNodeNump = 0 ;

	    /* Move the old file pointer to point to the latest entry.. */

	    while( currOldRADp && matchFound == FALSE )
	    {
		(*oldNodeNump)++ ;

		while( currNewRADp && matchFound == FALSE )
		{
		    (*newNodeNump)++ ;

		    if( !strcasecmp( currNewRADp -> rootedAtLine, currOldRADp -> rootedAtLine ) )
		    	matchFound = TRUE ;
	    	    else
	    	    	currNewRADp = currNewRADp -> nextRootedAtDetailsp ;

	    	} /* New file while loop */


		if( matchFound == FALSE )
		{
		    *newNodeNump = 0 ;	        /* Reset the node counter */
		    currNewRADp = baseNewRADp ; /* Reset the pointer */

	    	    currOldRADp = currOldRADp -> nextRootedAtDetailsp ;
		}
		
	    } /* Old file while loop */	
	}


	if( matchFound == TRUE )
	{
	    (*differOldRADpp) = currOldRADp ;
 	    (*differNewRADpp) = currNewRADp ;
	}

	return( matchFound ) ;
}


int
bufferRootedAt( DMfilep, lineBufferStructp, baseRADpp, currRADpp )
FILE * 	DMfilep ;
struct	lineBufferStruct * lineBufferStructp ;
struct 	rootedAtDetails ** baseRADpp, ** currRADpp ;
{
int	ret_val = OK ;

	/* Buffer the "rootedAt" line */
	if( (ret_val = getLine( DMfilep, lineBufferStructp, ROOTEDATLINETYPE ) ) == OK )
	{
	    if( !(*currRADpp) )
	    {
		if((*baseRADpp = *currRADpp = 
		    create_rootedAtDetails_node( ftell(DMfilep), 
		    	lineBufferStructp -> lineBufferp ) ) == NULLRAD )
		    ret_val = NOTOK ;
	    }
	    else
	    {
		if(((*currRADpp) -> nextRootedAtDetailsp = 
		    create_rootedAtDetails_node( ftell(DMfilep),
		    	lineBufferStructp -> lineBufferp ) ) == NULLRAD )
			ret_val = NOTOK ;
		else
			*currRADpp = (*currRADpp) -> nextRootedAtDetailsp ;
	    }

	    *(lineBufferStructp -> lineBufferp) = '\0' ;
	}

	return( ret_val ) ;
}


void
differEntries( oldDMfp, oldBufferStructp, newDMfp, newBufferStructp, deleteAttributeNotEntries )
FILE *	oldDMfp, * newDMfp ;
struct  lineBufferStruct * oldBufferStructp, * newBufferStructp ;
int	deleteAttributeNotEntries ;

{
struct 	rootedAtDetails * baseOldRADp = NULLRAD , * baseNewRADp = NULLRAD,
			* currOldRADp = NULLRAD , * currNewRADp = NULLRAD,
			* differOldRADp = NULLRAD , * differNewRADp = NULLRAD ;
int	oldNodeNum, newNodeNum ;
int	LastAdvancedFile = NEWFILE, NextAdvancedFile = OLDFILE ;
int 	getOldLineOK, getNewLineOK, getEntryOK ;
int	entryFound, AllEntriesLoopExit = FALSE, BothFilesInput = FALSE ;
int	lineType ;
long	preOldRootedAtOffset = ftell( oldDMfp ),
	preNewRootedAtOffset = ftell( newDMfp ) ;

	/* Perform an initial check that both files have the rootedAt before any attributes */
	if( (getOldLineOK = getLine( oldDMfp, oldBufferStructp, ROOTEDATORATTSLINETYPE )) != NOTOK )
	{
	    lineType = determineLineType( oldBufferStructp->lineBufferp ) ;

	    if( matchLineType( lineType, ROOTEDATLINETYPE ) == FALSE )
	    {
	    	fprintf( stderr, "FATAL ERROR : missing initial rootedAt in oldDMdiff file\n" ) ;
	    	return ;
	    }
	}

	if( (getNewLineOK = getLine( newDMfp, newBufferStructp, ROOTEDATORATTSLINETYPE )) != NOTOK )
	{
	    lineType = determineLineType( newBufferStructp->lineBufferp ) ;

	    if( matchLineType( lineType, ROOTEDATLINETYPE ) == FALSE )
	    {
	    	fprintf( stderr, "FATAL ERROR : missing initial rootedAt in newDMdiff file\n" ) ;
	    	return ;
	    }
	}

	/* Rewind the file offsets.. */
	(void)fseek( oldDMfp, preOldRootedAtOffset, 0 ) ;
	(void)fseek( newDMfp, preNewRootedAtOffset, 0 ) ;

	/* Reset the line buffers.. */
	*(oldBufferStructp -> lineBufferp) = '\0' ;
	*(newBufferStructp -> lineBufferp) = '\0' ;

	while( AllEntriesLoopExit == FALSE )
	{
	    /* Find the next "rootedAt" statement in both files */
	    if( (getOldLineOK = getLine( oldDMfp, oldBufferStructp, ROOTEDATLINETYPE )) != NOTOK &&
	        (getNewLineOK = getLine( newDMfp, newBufferStructp, ROOTEDATLINETYPE )) != NOTOK &&
  		(getOldLineOK == OK || getNewLineOK == OK ) )
	    {
		/* Check to see if the rootedAt statements correspond.. */
	    	if( getOldLineOK == OK && getNewLineOK == OK &&
		    !strcasecmp( oldBufferStructp -> lineBufferp, newBufferStructp -> lineBufferp ) )
		{
		    /* Output the rootedAt line.. */
		    printf( "\n\n%s\n", oldBufferStructp -> lineBufferp ) ;

		    *(oldBufferStructp -> lineBufferp) = '\0' ; /* Line Buffer contents now used */
		    *(newBufferStructp -> lineBufferp) = '\0' ;


		    if( differAttributes( oldDMfp, oldBufferStructp, 
			    			newDMfp, newBufferStructp, deleteAttributeNotEntries ) == NOTOK )
		    	AllEntriesLoopExit = TRUE ;
		}
		else
	    	{
		    /* If this is the first iteration then ensure that
		       both have supplied a rootedAt line ..           */
		    if( (baseOldRADp == NULLRAD && getOldLineOK != OK) ||
		        (baseNewRADp == NULLRAD && getNewLineOK != OK)   )
			AllEntriesLoopExit = TRUE ;
		     
   		    /* Buffer the two rootedAt lines */

		    if( AllEntriesLoopExit == FALSE && getOldLineOK == OK )
		    {
			if( baseOldRADp )
			{
			    if((currOldRADp -> nextRootedAtDetailsp = 
			    	create_rootedAtDetails_node( ftell(oldDMfp),
 				    oldBufferStructp -> lineBufferp) ) == NULLRAD )
				AllEntriesLoopExit = TRUE ;
			    else
				currOldRADp = currOldRADp -> nextRootedAtDetailsp ;
			}
			else
			{
			    if((baseOldRADp = currOldRADp = 
		    	    	create_rootedAtDetails_node( ftell(oldDMfp),
 				    oldBufferStructp -> lineBufferp)) == NULLRAD )
				AllEntriesLoopExit = TRUE ;
			}
			
		    	*(oldBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */
		    }

		    if( AllEntriesLoopExit == FALSE && getNewLineOK == OK )
		    {
			if( baseNewRADp )
			{
			    if((currNewRADp -> nextRootedAtDetailsp = 
			    	create_rootedAtDetails_node( ftell(newDMfp),
 				    newBufferStructp -> lineBufferp) ) == NULLRAD )
				AllEntriesLoopExit = TRUE ;
			    else
				currNewRADp = currNewRADp -> nextRootedAtDetailsp ;
			}
			else
			{
			    if((baseNewRADp = currNewRADp = 
		    	    	create_rootedAtDetails_node( ftell(newDMfp), 
				    newBufferStructp -> lineBufferp )) == NULLRAD )
				AllEntriesLoopExit = TRUE ;
			}
			
		    	*(newBufferStructp -> lineBufferp) = '\0' ; /* line has now been used */
		    }

		    entryFound = FALSE ;


		    if( getOldLineOK == OK && getNewLineOK == OK )
		    {
		    	BothFilesInput = TRUE ;
		    }
		    else
		    {
		    	BothFilesInput = FALSE ;

		    	if( getOldLineOK == OK )
			    LastAdvancedFile = OLDFILE ;

			else if( getNewLineOK == OK )
		    	    LastAdvancedFile = NEWFILE ;
		    }

		    

		    while( AllEntriesLoopExit == FALSE && entryFound == FALSE &&
			    		(getOldLineOK == OK || getNewLineOK == OK) )
	    	    {
			/* Check for a match between the old and new file lists.. */

			if( findRootedAtMatch( baseOldRADp, baseNewRADp,
			      &oldNodeNum, &newNodeNum, &differOldRADp, &differNewRADp,
	   			(BothFilesInput == TRUE) ? BOTHFILES : LastAdvancedFile ) 
					== TRUE )
		   	{
#ifdef DEBUG
fprintf( stderr, "ATTR MATCH %d %d\n", oldNodeNum, newNodeNum );
#endif
			    if( differAttributesFromRADnode( oldDMfp, differOldRADp, oldBufferStructp,
					newDMfp, differNewRADp, newBufferStructp, deleteAttributeNotEntries ) == NOTOK )
			    	AllEntriesLoopExit = TRUE ;
			    else
			    {
			    	entryFound = TRUE ;

				/* Unlink the matched node pair.. */
#ifdef DEBUG1
fprintf( stderr, "Freeing old node %d\n", oldNodeNum );

fprintf( stderr, "Freeing new node %d\n", newNodeNum ) ;

fprintf( stderr, "CURRENR BUFFER ON FREE[%s]", oldBufferStructp -> lineBufferp) ; /* Line Buffer contents now processed */
#endif
				currOldRADp = freeRootAtDetails_node( &baseOldRADp, oldNodeNum ) ;
				currNewRADp = freeRootAtDetails_node( &baseNewRADp, newNodeNum ) ;

				/* Ensure that there will be something to match
				   against next time */
		    		if( (currOldRADp == NULLRAD && getOldLineOK != OK) ||
			    	    (currNewRADp == NULLRAD && getNewLineOK != OK)   )
				    AllEntriesLoopExit ;
			    }


			    continue ;
			}
			
			BothFilesInput = FALSE ;

			/* Determine which file to advance next */

		    	if( LastAdvancedFile == OLDFILE && getNewLineOK == OK )
		    	    NextAdvancedFile =  NEWFILE ;

		    	if( LastAdvancedFile == NEWFILE && getOldLineOK == OK )
		    	    NextAdvancedFile =  OLDFILE ;


			/* Try advancing one file to the next rootedAt 
			   statement (new file first then the old file) */

		    	if( NextAdvancedFile == NEWFILE )
		    	{
			    /* Advance new file by one entry to the next "rootedAt".. */
			    getNewLineOK = getLine( newDMfp, newBufferStructp, ROOTEDATLINETYPE) ;

			    if( getNewLineOK == OK )
			    {
			    	if((currNewRADp -> nextRootedAtDetailsp = 
				    create_rootedAtDetails_node( ftell(newDMfp), 
 				    	newBufferStructp -> lineBufferp) ) == NULLRAD )
				    AllEntriesLoopExit = TRUE ;
			    	else
				    currNewRADp = currNewRADp -> nextRootedAtDetailsp ;

				*(newBufferStructp -> lineBufferp) = '\0' ; /* Line Buffer contents now processed */
			    }


			    LastAdvancedFile = NEWFILE ;
			}
			else
		    	{
			    /* Advance old file by one entry to the next "rootedAt".. */
			    getOldLineOK = getLine( oldDMfp, oldBufferStructp, ROOTEDATLINETYPE) ;

			    if( getOldLineOK == OK )
			    {
			    	if((currOldRADp -> nextRootedAtDetailsp = 
				    create_rootedAtDetails_node( ftell(oldDMfp),
 				    	oldBufferStructp -> lineBufferp) ) == NULLRAD )
				    AllEntriesLoopExit = TRUE ;
			    	else
				    currOldRADp = currOldRADp -> nextRootedAtDetailsp ;

				*(oldBufferStructp -> lineBufferp) = '\0' ; /* Line Buffer contents now processed */
			    }
			    
			    LastAdvancedFile = OLDFILE ;
		    	}

			if( getOldLineOK == NOTOK || getNewLineOK == NOTOK )
			    AllEntriesLoopExit = TRUE ;
			

#ifdef DEBUG1
fprintf ( stderr, ">>>> Entryfound %d getOld %d getNew %d Next %d Last %d Exit %d\n",
		    		entryFound, getOldLineOK, getNewLineOK,
				NextAdvancedFile, LastAdvancedFile, AllEntriesLoopExit );
#endif
		    } /* while: current rootedAt entry */
		}
	    }
	    else
	    {
	    	/* Just in case the last successful read actually set up TWO 
		   pairwise matches check all possible pairwise matches ..*/
		   
		if( findRootedAtMatch( baseOldRADp, baseNewRADp,
			&oldNodeNum, &newNodeNum, &differOldRADp, &differNewRADp,
			  BOTHFILES) == TRUE )
		{
#ifdef DEBUG
fprintf( stderr, "ATTR MATCH %d %d\n", oldNodeNum, newNodeNum );
#endif
		    if( differAttributesFromRADnode( oldDMfp, differOldRADp, oldBufferStructp,
					newDMfp, differNewRADp, newBufferStructp, deleteAttributeNotEntries ) != NOTOK )
		    {
			/* Unlink the matched node pair.. */
#ifdef DEBUG1
fprintf( stderr, "Freeing old node %d\n", oldNodeNum );

fprintf( stderr, "Freeing new node %d\n", newNodeNum ) ;

fprintf( stderr, "CURRENR BUFFER ON FREE[%s]", oldBufferStructp -> lineBufferp) ; /* Line Buffer contents now processed */
#endif
			currOldRADp = freeRootAtDetails_node( &baseOldRADp, oldNodeNum ) ;
			currNewRADp = freeRootAtDetails_node( &baseNewRADp, newNodeNum ) ;

		    }
		}
		
		/* Neither file have entries remaining */
	    	AllEntriesLoopExit = TRUE ;
	    }
	    
	} /* all rootedAt entries loop */

#ifdef DEBUG2
fprintf ( stderr, "<>>>> Entryfound %d getOld %d getNew %d Next %d Last %d Exit %d\n",
		    		entryFound, getOldLineOK, getNewLineOK,
				NextAdvancedFile, LastAdvancedFile, AllEntriesLoopExit );
#endif


	/* Read in any remaining entries (only one file can be outstanding) */

	getEntryOK = OK ;

	if( getOldLineOK == OK )
	{
	    while( getEntryOK == OK )
	    {	    
		getEntryOK = bufferRootedAt( oldDMfp, oldBufferStructp,
							&baseOldRADp, &currOldRADp ) ;
	    }
	}
	else if( getNewLineOK == OK )
	{
	    while( getEntryOK == OK )
	    {
		getEntryOK = bufferRootedAt( newDMfp, newBufferStructp,
							&baseNewRADp, &currNewRADp ) ;

	    }
	}


	/* Process any remaining entries requiring deletion */
	(void)deleteEntriesFromRADlist( oldDMfp, baseOldRADp, 
			    		oldBufferStructp, ALLNODES, deleteAttributeNotEntries ) ;

	/* Process any remaining entries requiring addition */
	(void)addEntriesFromRADlist( newDMfp, baseNewRADp, 
			    		newBufferStructp, ALLNODES ) ;
					
	    

	/* Free any remaining linked list storage */
	freeRootAtDetails_list( baseOldRADp ) ;
	freeRootAtDetails_list( baseNewRADp ) ;


	return ;

} /* differEntries() */



void
outputHeader()
{
	printf( "# Notation:-\n" ) ;

	printf( "# %cXXX indicates that XXX is an       RDN       that is to be ADDED\n", ADDRDNCHAR );
	printf( "# %cXXX indicates that XXX is an       RDN       that is to be DELETED\n", DELETERDNCHAR );

	printf( "# %cXXX indicates that XXX is an    ATTRIBUTE    that is to be ADDED\n", ADDATTRCHAR );
	printf( "# %cXXX indicates that XXX is an    ATTRIBUTE    that is to be DELETED\n", DELETEATTRCHAR );

	printf( "# %cXXX indicates that XXX is an ATTRIBUTE VALUE that is to be ADDED\n", ADDVALUECHAR );
	printf( "# %cXXX indicates that XXX is an ATTRIBUTE VALUE that is to be DELETED\n", DELETEVALUECHAR );

} /* outputHeader() */




void	
produceDMdiff( oldDMfp, newDMfp, deleteAttributeNotEntries )
FILE *  oldDMfp, * newDMfp ;
int	deleteAttributeNotEntries ;
{
struct  lineBufferStruct oldBufferStruct, newBufferStruct ;
char *	oldSourceNamecp, * newSourceNamecp ;

	/* .. these line buffers are passed down the procedure stack to
	   allow for read lines that were not processed to be passed on
	   to the next stage. */


	if( (oldBufferStruct.lineBufferp = malloc( INITLINEBUFLEN )) == NULLCP ||
	    (newBufferStruct.lineBufferp = malloc( INITLINEBUFLEN )) == NULLCP   )
	{
	    fprintf( stderr, "Memory allocation failure when creating the line buffers\n" );

	    return ;
	}
	else
	{
	    oldBufferStruct.bufferSize = INITLINEBUFLEN ;
	    newBufferStruct.bufferSize = INITLINEBUFLEN ;
	}
	

	*(oldBufferStruct.lineBufferp) = *(newBufferStruct.lineBufferp) = '\0' ;


	if( getNextLine( oldDMfp, oldBufferStruct.lineBufferp, SOURCELINETYPE ) == NOTOK ||
	    getNextLine( newDMfp, newBufferStruct.lineBufferp, SOURCELINETYPE ) == NOTOK   )
	{
	    fprintf( stderr, "Couldnot find source line(s)\n" ) ;
	}
	else
	{
	    /* Check that the sources do indeed match */

	    if((oldSourceNamecp = strDupTidyStringRemLineConts( oldBufferStruct.lineBufferp )) == NULLCP ||
	       (newSourceNamecp = strDupTidyStringRemLineConts( newBufferStruct.lineBufferp )) == NULLCP   ) {
		fprintf( stderr, "FATAL ERROR: Memory allocation problem when parsing source statements\n" ) ;

	    } else if( strlen( oldSourceNamecp ) <= strlen( "source=" ) ) {
		fprintf( stderr, "Invalid source\n%s\n",
	    	    oldBufferStruct.lineBufferp ) ;

	    } else if( strlen( newSourceNamecp ) <= strlen( "source=" ) ) {
		fprintf( stderr, "Invalid source\n%s\n",
	    	    newBufferStruct.lineBufferp ) ;

	    } else if( strcasecmp( oldSourceNamecp, newSourceNamecp ) ) {
		fprintf( stderr, "Sources differ\n%s\n%s\n",
	    	    oldBufferStruct.lineBufferp, newBufferStruct.lineBufferp ) ;
	    }
	    else
	    {
		printf( "newdmformat [%s]\n", Version );
		
		/* Output the source line */
		printf( "\n%s\n\n", oldBufferStruct.lineBufferp );

		*(oldBufferStruct.lineBufferp) = *(newBufferStruct.lineBufferp) = '\0' ;

		/* Output the header message */
		outputHeader() ;
		
		/* Check for differences between the entries.. */
		differEntries( oldDMfp, &oldBufferStruct, newDMfp, &newBufferStruct, deleteAttributeNotEntries ) ;
	    }

	    (void)free( oldSourceNamecp ) ; (void)free( newSourceNamecp ) ;
	}

	free( oldBufferStruct.lineBufferp ) ; free( newBufferStruct.lineBufferp ) ; 

	return ;
}


main( argc, argv )
int 	argc ;
char *	argv[] ;
{
FILE *	oldDMfp, * newDMfp ;
int	deleteAttributeNotEntries, currarg ;
char	oldfilename[ SMALLBUFLEN ], newfilename[ SMALLBUFLEN ] ;

	oldfilename[ 0 ] = newfilename[ 0 ] = '\0' ;

	/* Check the number of parameters */

	if( argc < 3 || argc > 4 )
	{
	    fprintf( stderr, "Please enter\nnewdmdiff <oldDMfilename> <newDMfilename>\n" ) ;
	    fprintf( stderr, "OR for secondary data sources :-\nnewdmdiff -s  <oldDMfilename> <newDMfilename>\n" ) ;
	    fprintf( stderr, "OR also for secondary data sources :-\nnewdmdiff -so <oldDMfilename> <newDMfilename>\n" ) ;
	    exit( -1 );
	}

	deleteAttributeNotEntries = DELENTRIESNOTATTRS ;

	for( currarg = 1 ; currarg < argc ; currarg++ )
	{
	    if( *(argv[ currarg ]) == '-' )
	    {
	    	/* A command parameter.. */
	    	if( tolower( *(argv[ currarg ] + 1)) == 's' )
		{
		    if( tolower( *(argv[ currarg ] + 2 )) == 'o' )
		    {
		    	deleteAttributeNotEntries = DELATTRSNOTENTRIES ;
		    }
		    else
		    {
			deleteAttributeNotEntries = DELATTRSNOTOBJCLSORENTS ;
		    }
		}
		else
		{
		    fprintf( stderr, "Unknown flag: -%c\n", *(argv[ currarg ] + 1) ) ;
		    exit( -1 ) ;
		}
	    }
	    else
	    {
	    	/* A file name.. */
		if( oldfilename[ 0 ] == '\0' )
		    strcpy( oldfilename, argv[ currarg ] ) ;
		else
		    strcpy( newfilename, argv[ currarg ] ) ;
	    }
	}

	/* Check that both file names have been input.. */
	if( oldfilename[ 0 ] == '\0' || newfilename[ 0 ] == '\0' )
	{
	    fprintf( stderr, "Please enter\nnewdmdiff <oldDMfilename> <newDMfilename>\n" ) ;
	    fprintf( stderr, "OR for secondary data sources :-\nnewdmdiff -s  <oldDMfilename> <newDMfilename>\n" ) ;
	    fprintf( stderr, "OR also for secondary data sources :-\nnewdmdiff -so <oldDMfilename> <newDMfilename>\n" ) ;
	    exit( -1 );
	}

	/* Check that both files can be opened for reading.. */

	if( (oldDMfp = fopen( oldfilename, "r" )) == (FILE *)NULL )
	{
	    perror( "oldDMdiff file name could not be opened" );
 	    exit( -1 );
	}

	if( (newDMfp = fopen( newfilename, "r" )) == (FILE *)NULL )
	{
	    perror( "newDMdiff file name is opened" );
 	    exit( -1 );
	}

#ifndef PPANDDLUNAVAILABLE
	sys_init(argv[0]);
	or_myinit();
#endif
	/* Load syntax handlers .. */
	quipu_syntaxes();

#ifndef PPANDDLUNAVAILABLE
	pp_syntaxes();
#endif

	/* Initialise the DSAP library (to indicate the oidtable directory) */
	dsap_init((int *)NULL, (char ***)NULL);

	
#ifndef PPANDDLUNAVAILABLE
  	pp_quipu_run();
#endif

	produceDMdiff( oldDMfp, newDMfp, deleteAttributeNotEntries );


	if( fclose( oldDMfp ) || fclose( newDMfp ) )
	{
	    perror( "File closure problem" );
	    exit( -1 ) ;
	}
	
	fprintf( stderr, "Total RDNs processed = %d\n", totalRDNs ) ;

	exit( 0 );
}
